# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

import math
import time
import logging
import itertools
import argparse
import os, sys

import numpy as np
from numpy import corrcoef

from tvb.simulator.lab import *
from tvb.dsl_cuda import LEMS2CUDA


class TVB_test:

	def __init__(self):
		self.args = self.parse_args()
		self.sim_length = self.args.n_time # 400
		self.g = np.array([1.0])
		self.s = np.array([1.0])
		self.dt = 0.1
		self.period = 10.0
		self.omega = 60.0 * 2.0 * math.pi / 1e3
		(self.connectivity, self.coupling) = self.tvb_connectivity(self.s, self.g, self.dt)
		self.integrator = integrators.EulerDeterministic(dt=self.dt)
		self.weights = self.SC = self.connectivity.weights
		self.lengths = self.connectivity.tract_lengths
		self.n_nodes = self.weights.shape[0]
		self.tavg_period = 10.0
		self.nstep = self.args.n_time  # 4s
		self.n_inner_steps = int(self.tavg_period / self.dt)
		self.nc = self.args.n_coupling
		self.ns = self.args.n_speed
		self.couplings, self.speeds = self.setup_params(self.nc, self.ns)
		self.params = self.expand_params(self.couplings, self.speeds)
		self.n_work_items, self.n_params = self.params.shape
		self.min_speed = self.speeds.min()
		self.buf_len_ = ((self.lengths / self.min_speed / self.dt).astype('i').max() + 1)
		self.buf_len = 2 ** np.argwhere(2 ** np.r_[:30] > self.buf_len_)[0][0]  # use next power of 2
		self.states = 1

	def tvb_connectivity(self, speed, global_coupling, dt=0.1):
		white_matter = connectivity.Connectivity.from_file(source_file="connectivity_68.zip")
		white_matter.configure()
		white_matter.speed = np.array([speed])
		white_matter_coupling = coupling.Linear(a=global_coupling)
		return white_matter, white_matter_coupling

	def parse_args(self):  # {{{
		parser = argparse.ArgumentParser(description='Run parameter sweep.')
		parser.add_argument('-c', '--n_coupling', help='num grid points for coupling parameter', default=32, type=int)
		parser.add_argument('-s', '--n_speed', help='num grid points for speed parameter', default=32, type=int)
		parser.add_argument('-t', '--test', help='check results', action='store_true')
		parser.add_argument('-n', '--n_time', help='number of time steps to do (default 400)', type=int, default=400)
		parser.add_argument('-v', '--verbose', help='increase logging verbosity', action='store_true', default='-v')
		# parser.add_argument('-p', '--no_progress_bar', help='suppress progress bar', action='store_false')
		# parser.add_argument('--caching',
		# 					choices=['none', 'shared', 'shared_sync', 'shuffle'],
		# 					help="caching strategy for j_node loop (default shuffle)",
		# 					default='none'
		# 					)
		parser.add_argument('--node_threads', default=1, type=int)
		parser.add_argument('--model',
							choices=['Rwongwang', 'Kuramoto', 'Epileptor', 'Oscillator', \
									 'Oscillatorref', 'Kuramotoref', 'Rwongwangref', 'Epileptorref'],
							help="neural mass model to be used during the simulation",
							default='Oscillator'
							)
		parser.add_argument('--lineinfo', default=True, action='store_true')

		parser.add_argument('--filename', default="kuramoto_network.c", type=str,
							help="Filename to use as GPU kernel definition")

		parser.add_argument('-b', '--bench', default="regular", type=str, help="What to bench: regular, numba, cuda")

		parser.add_argument('-bx', '--blockszx', default="32", type=int, help="Enter block size x")
		parser.add_argument('-by', '--blockszy', default="32", type=int, help="Enter block size y")

		parser.add_argument('-val', '--validate', default=False, help="Enable validation to refmodels")

		args = parser.parse_args()
		return args

	def expand_params(self, couplings, speeds):  # {{{
		# the params array is transformed into a 2d array
		# by first creating tuples of (speed, coup) and arrayfying then
		# pycuda (check) threats them as flattenened arrays but numba needs 2d indexing
		params = itertools.product(speeds, couplings)
		params = np.array([vals for vals in params], np.float32)
		return params  # }}}

	def setup_params(self, nc, ns):  # {{{
		# the correctness checks at the end of the simulation
		# are matched to these parameter values, for the moment
		couplings = np.logspace(1.6, 3.0, nc)
		speeds = np.logspace(0.0, 2.0, ns)
		return couplings, speeds  # }}}

	# Todo: check if this function work. derr_speed > 500 and derr_coupl < -1500 evaluate to false for pyCuda runs
	def check_results(self, n_nodes, n_work_items, tavg, weights, speeds, couplings, logger, args):
		r, c = np.triu_indices(n_nodes, 1)
		win_size = args.n_time  # 4s? orig 200 # 2s?
		win_tavg = tavg.reshape((-1, win_size) + tavg.shape[1:])
		err = np.zeros((len(win_tavg), n_work_items))
		logger.info('err.shape %s', err.shape)
		# TODO do cov/corr in kernel
		for i, tavg_ in enumerate(win_tavg):
			for j in range(n_work_items):
				fc = np.corrcoef(tavg_[:, :, j].T)
				# err[i, j] = ((fc[r, c] - weights[r, c])**2).sum()   weights is 1 dim array
				# logger.info('fc[r, c].shape %s, weights[r].shape %s', fc[r, c].shape, weights[r].shape)
				err[i, j] = ((fc[r, c] - weights[r, c]) ** 2).sum()
		# look at 2nd 2s window (converges quickly)
		err_ = err[-1].reshape((speeds.size, couplings.size))
		# change on fc-sc metric wrt. speed & coupling strength
		derr_speed = np.diff(err_.mean(axis=1)).sum()
		derr_coupl = np.diff(err_.mean(axis=0)).sum()
		logger.info('derr_speed=%f, derr_coupl=%f', derr_speed, derr_coupl)
		# if args.dataset == 'hcp':
		assert derr_speed > 350.0
		assert derr_coupl < -500.0
		# if args.dataset == 'sep':
		# 	assert derr_speed > 5e4
		# 	assert derr_coupl > 1e4

		logger.info('result OK')

	def start_cuda(self, logger):
		# logger.info('start Cuda run')
		import pycuda.autoinit
		from tvb.dsl_cuda.run.cuda_run import CudaRun
		cudarun = CudaRun()
		tavg_data = cudarun.run_simulation(self.weights, self.lengths, self.params, self.speeds, logger,
										   self.args, self.n_nodes, self.n_work_items, self.n_params, self.nstep,
										   self.n_inner_steps, self.buf_len, self.states, self.dt, self.min_speed)

		# Todo: fix this for cuda
		# self.check_results(self.n_nodes, self.n_work_items, tavg_data, self.weights, self.speeds, self.couplings, logger, self.args)
		return tavg_data

	def set_CUDAmodel_dir(self):
		# self.args.filename = os.path.join(os.path.join(os.getcwd(), os.pardir), 'CUDAmodels', self.args.model.lower() + '.c')
		self.args.filename = os.path.join((os.path.dirname(os.path.abspath(__file__))), os.pardir, 'CUDAmodels',
								 self.args.model.lower() + '.c')

	def set_states(self):
		if 'kuramoto' in self.args.model.lower():
			self.states = 1
		elif 'oscillator' in self.args.model.lower():
			self.states = 2
		elif 'wongwang' in self.args.model.lower():
			self.states = 2
		elif 'montbrio' in self.args.model.lower():
			self.states = 2
		elif 'epileptor' in self.args.model.lower():
			self.states = 6

	def compare_with_ref(self, logger, tavg0):
		self.args.model = self.args.model + 'ref'
		self.set_CUDAmodel_dir()
		tavg1 = self.start_cuda(logger)

		# compare output to check if same as template
		comparison = (tavg0.ravel() == tavg1.ravel())
		logger.info('Templated version is similar to original %d:', comparison.all())
		logger.info('Correlation coefficient: %f', corrcoef(tavg0.ravel(), tavg1.ravel())[0, 1])

	def startsim(self):

		tic = time.time()
		logging.basicConfig(level=logging.DEBUG if self.args.verbose else logging.INFO)
		logger = logging.getLogger('[TVB_CUDA]')

		# logger.info('dt %f', self.dt)
		# logger.info('nstep %d', self.nstep)
		# logger.info('n_inner_steps %f', self.n_inner_steps)
		# if self.args.test and self.args.n_time % 200:
		# 	logger.warning('rerun w/ a multiple of 200 time steps (-n 200, -n 400, etc) for testing')  # }}}
		#
		# # setup data
		# logger.info('weights.shape %s', self.weights.shape)
		# logger.info('lengths.shape %s', self.lengths.shape)
		# logger.info('n_nodes %d', self.n_nodes)
		#
		# # couplings and speeds are not derived from the regular TVB connection setup routine.
		# # these parameters are swooped every GPU spawn
		# logger.info('single connectome, %d x %d parameter space', self.ns, self.nc)
		# logger.info('%d total num threads', self.ns * self.nc)
		# logger.info('min_speed %f', self.min_speed)
		# logger.info('real buf_len %d, using power of 2 %d', self.buf_len_, self.buf_len)

		self.set_CUDAmodel_dir()

		self.set_states()

		# logger.info('number of states %d', self.states)
		# logger.info('filename %s', self.args.filename)
		# logger.info('model %s', self.args.model)

		tac = time.time()
		# logger.info("Setup in: {}".format(tac - tic))

		tavg0 = self.start_cuda(logger)
		toc = time.time()

		if (self.args.validate):
			self.compare_with_ref(logger, tavg0)

		toc = time.time()
		elapsed = toc - tic
		logger.info('Finished python simulation successfully in: %0.3f', elapsed)
		logger.info('%0.3f M step/s', 1e-6 * self.nstep * self.n_inner_steps * self.n_work_items / elapsed)
		# logger.info('finished')


if __name__ == '__main__':

	np.random.seed(79)

	example = TVB_test()

	# start templating the model specified on cli
	here = os.path.abspath(os.path.dirname(__file__))
	LEMS2CUDA.cuda_templating(example.args.model, os.path.join(here, '..', 'XMLmodels'))

	# start simulation with templated model
	example.startsim()



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Tests stimulation when sub-cortical regions are present (in the connectivity matrix).

.. moduleauthor:: Borana Dollomaja <borana.dollomaja@univ-amu.fr>
"""

import numpy

from tvb.simulator.lab import *
from tvb.tests.library.base_testcase import BaseTestCase


class PulseStimulation(object):
    """
    Stimulus test class
    """
    def __init__(self, con):
        # Configure stimulus spatial pattern.
        stim_node = [0] # target zone where we want to stimulate
        stim_weights = numpy.zeros((con.number_of_regions, ))
        stim_weights[stim_node] = numpy.array([2.0])

        # Configure stimulus temporal profile.
        eqn_t = equations.PulseTrain()
        eqn_t.parameters['onset'] = 2e3       # onset time [ms]
        eqn_t.parameters['T'] = 3000.0        # pulse repetition period [ms]
        eqn_t.parameters['tau'] = 100.0       # pulse duration [ms]

        # Configure Stimuli object.
        self.stimulus = patterns.StimuliRegion(temporal = eqn_t,
                                        connectivity = con, 
                                        weight =stim_weights)

class SetUpSimulator(object):
    """
    Stimulation test class
    """
    def __init__(self):
        """
        Initialize the structural information, coupling function, integrator, monitors, surface and stimulation.
        """
        # Connectome
        con = connectivity.Connectivity.from_file("connectivity_192.zip")
        con.configure()

        # Surface and local connectivity kernel
        surf = cortex.Cortex.from_file() #Initialise a surface
        surf.local_connectivity = local_connectivity.LocalConnectivity.from_file()
        surf.configure()

        # Model
        oscilator = models.Generic2dOscillator()

        self.sim = simulator.Simulator(
            conduction_speed=1.0,
            coupling= coupling.Difference(a=numpy.array([0.01])),
            surface=surf,
            stimulus=PulseStimulation(con).stimulus,
            integrator=integrators.Identity(dt=1.0),
            simulation_length=10.0,
            connectivity=con,
            model=oscilator,
            monitors=(monitors.Raw(),)
        )
        self.sim.configure()

    def run_simulation(self):
        return self.sim.run(simulation_length = 10)

class TestStimulation(BaseTestCase):
    def test_stimulation(self):
        model = SetUpSimulator()
        result = model.run_simulation()
        assert result is not None


# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Filler analyzer: Takes a TimeSeries object and returns a Float.

.. moduleauthor:: Paula Sanz Leon <pau.sleon@gmail.com>

"""

import cmath
import numpy
from tvb.basic.logger.builder import get_logger


"""
Return the Kuramoto synchronization index. 
    
Useful metric for a parameter analysis when the collective brain dynamics
represent coupled oscillatory processes.
    
The *order* parameters are :math:`r` and :math:`Psi`.
    
.. math::
    r e^{i * \\psi} = \\frac{1}{N}\\,\\sum_{k=1}^N(e^{i*\\theta_k})
    
The first is the phase coherence of the population of oscillators (KSI) 
and the second is the average phase.
    
When :math:`r=0` means 0 coherence among oscillators.
    
    
Input:
TimeSeries DataType
    
Output: 
Float
    
This is a crude indicator of synchronization among nodes over the entire network.

#NOTE: For the time being it is meant to be another global metric.
However, it should be consider to have a sort of TimeSeriesDatatype for this
analyzer.
    
"""


log = get_logger(__name__)


def compute_kuramoto_index_metric(params):
    """
    # type: dict(TimeSeries)  -> float
    Kuramoto Synchronization Index

    Parameters
    ----------
    params : a dictionary containing
        time_series : TimeSeries
        Input time series for which the metric will be computed.
    """

    time_series = params['time_series']
    if time_series.data.shape[1] < 2:
        msg = " The number of state variables should be at least 2."
        log.error(msg)
        raise Exception(msg)
                
    # TODO: Should be computed for each possible combination of var, mode
    #      for var, mode in itertools.product(range(self.time_series.data.shape[1]),
    #                                         range(self.time_series.data.shape[3])):
        
    # TODO: Generalise. The Kuramoto order parameter is computed over sliding
    #      time windows and then normalised

    theta_sum = numpy.sum(numpy.exp(0.0 + 1j * (numpy.vectorize(cmath.polar)
                (numpy.vectorize(complex)(time_series.data[:, 0, :, 0],
                    time_series.data[:, 1, :, 0]))[1])), axis=1)
                     
    result = numpy.vectorize(cmath.polar)(theta_sum / time_series.data.shape[2])

    return result[0].mean()



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Singleton logging builder.


.. moduleauthor:: Calin Pavel <calin.pavel@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""
import importlib
import os
import weakref
import logging
import logging.config
from tvb.basic.profile import TvbProfile
from tvb.basic.config.profile_settings import MATLABLibraryProfile


class LoggerBuilder(object):
    """
    Class taking care of uniform Python logger initialization. 
    It uses the Python native logging package. 
    It's purpose is just to offer a common mechanism for initializing all modules in a package.
    """

    def __init__(self, config_root):
        """
        Prepare Python logger based on a configuration file.
        :param: config_root - current package to configure logger for it.

        """
        if not isinstance(TvbProfile.current, MATLABLibraryProfile):
            config_file_name = TvbProfile.current.LOGGER_CONFIG_FILE_NAME
            package = importlib.import_module(config_root)
            package_path = package.__path__[0]
            # Specify logging configuration file for current package.
            logging.config.fileConfig(os.path.join(package_path, config_file_name), disable_existing_loggers=False)
        else:
            logging.basicConfig(level=logging.DEBUG)
        self._loggers = weakref.WeakValueDictionary()

    def build_logger(self, parent_module):
        """
        Build a logger instance and return it
        """
        self._loggers[parent_module] = logger = logging.getLogger(parent_module)
        return logger

    def set_loggers_level(self, level):
        for logger in self._loggers.values():
            logger.setLevel(level)


# We make sure a single instance of logger-builder is created.
if "GLOBAL_LOGGER_BUILDER" not in globals():

    if TvbProfile.is_library_mode():
        GLOBAL_LOGGER_BUILDER = LoggerBuilder('tvb.basic.logger')
    else:
        GLOBAL_LOGGER_BUILDER = LoggerBuilder('tvb.config.logger')


def get_logger(parent_module=''):
    """
    Function to retrieve a new Python logger instance for current module.
    
    :param parent_module: module name for which to create logger.
    """
    return GLOBAL_LOGGER_BUILDER.build_logger(parent_module)



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Dionysios Perdikis <dionperd@gmail.com>
.. moduleauthor:: Gabriel Florea <gabriel.florea@codemart.ro>
"""
import os
from datetime import datetime

from tvb.basic.profile import TvbProfile


class FiguresConfig(object):
    VERY_LARGE_SIZE = (40, 20)
    VERY_LARGE_PORTRAIT = (30, 50)
    SUPER_LARGE_SIZE = (80, 40)
    LARGE_SIZE = (20, 15)
    SMALL_SIZE = (15, 10)
    NOTEBOOK_SIZE = (20, 10)
    DEFAULT_SIZE = SMALL_SIZE
    FIG_FORMAT = 'png'
    SAVE_FLAG = True
    SHOW_FLAG = False
    MOUSE_HOOVER = False
    MATPLOTLIB_BACKEND = "Agg"  # "Qt4Agg"
    WEIGHTS_NORM_PERCENT = 99
    FONTSIZE = 10
    SMALL_FONTSIZE = 8
    LARGE_FONTSIZE = 12

    def largest_size(self):
        import sys
        if 'IPython' not in sys.modules:
            return self.LARGE_SIZE
        from IPython import get_ipython
        if getattr(get_ipython(), 'kernel', None) is not None:
            return self.NOTEBOOK_SIZE
        else:
            return self.LARGE_SIZE

    def __init__(self, out_base=None, separate_by_run=False):
        """
        :param out_base: Base folder where figures should be kept
        :param separate_by_run: Set TRUE, when you want figures to be in different files / each run
        """
        self._out_base = out_base or TvbProfile.current.TVB_STORAGE or os.path.join(os.getcwd(), "outputs")
        self._separate_by_run = separate_by_run

    @property
    def FOLDER_FIGURES(self):
        folder = os.path.join(self._out_base, "figs")
        if self._separate_by_run:
            folder = folder + datetime.strftime(datetime.now(), '%Y-%m-%d_%H-%M')
        if not (os.path.isdir(folder)):
            os.makedirs(folder)
        return folder


CONFIGURED = FiguresConfig()



"""
Expression parser

@author: Gautham Ganapathy
@organization: LEMS (http://neuroml.org/lems/, https://github.com/organizations/LEMS)
@contact: gautham@lisphacker.org

MV: added pow(f,l) function for c translations
"""

from lems.base.base import LEMSBase
from lems.base.stack import Stack

known_functions = ['exp', 'log', 'sqrt', 'sin', 'cos', 'tan', 'sinh', 'cosh', 'tanh', 'abs', 'ceil', 'factorial', 'random', 'H', 'pow', 'powf', 'powl']

class ExprNode(LEMSBase):
    """
    Base class for a node in the expression parse tree.
    """

    OP = 1
    VALUE = 2
    FUNC1 = 3

    def __init__(self, type):
        """
        Constructor.

        @param type: Node type
        @type type: enum(ExprNode.OP, ExprNode.VALUE)
        """

        self.type = type
        """ Node type.
        @type: enum(ExprNode.OP, ExprNode.VALUE) """

class ValueNode(ExprNode):
    """
    Value node in an expression parse tree. This will always be a leaf node.
    """

    def __init__(self, value):
        """
        Constructor.

        @param value: Value to be stored in this node.
        @type value: string
        """

        ExprNode.__init__(self, ExprNode.VALUE)
        self.value = value
        """ Value to be stored in this node.
        @type: string """
        
    def clean_up(self):
        """
        To make sure an integer is returned as a float. No division by integers!!
        """
        try:
            return str(float(self.value)) 
        except ValueError:
            return self.value
                

    def __str__(self):
        """
        Generates a string representation of this node.
        """
        return "{" + self.clean_up() + "}"
    
    def __repr__(self):
        return self.__str__()
    
    def to_python_expr(self):
        return self.clean_up()

class OpNode(ExprNode):
    """
    Operation node in an expression parse tree. This will always be a
    non-leaf node.
    """

    def __init__(self, op, left, right):
        """
        Constructor.

        @param op: Operation to be stored in this node.
        @type op: string

        @param left: Left operand.
        @type left: lems.parser.expr.ExprNode

        @param right: Right operand.
        @type right: lems.parser.expr.ExprNode
        """

        ExprNode.__init__(self, ExprNode.OP)

        self.op = op
        """ Operation stored in this node.
        @type: string """

        self.left = left
        """ Left operand.
        @type: lems.parser.expr.ExprNode """

        self.right = right
        """ Right operand.
        @type: lems.parser.expr.ExprNode """

    def __str__(self):
        """
        Generates a string representation of this node.
        """

        return '({0} {1} {2})'.format(self.op,
                                      str(self.left),
                                      str(self.right))
                                      
    def __repr__(self):
        return self.__str__()
    
    def to_python_expr(self):
        return '({0} {1} {2})'.format(self.left.to_python_expr(),
                                      self.op,
                                      self.right.to_python_expr())

class Func1Node(ExprNode):
    """
    Unary function node in an expression parse tree. This will always be a
    non-leaf node.
    """

    def __init__(self, func, param):
        """
        Constructor.

        @param func: Function to be stored in this node.
        @type func: string

        @param param: Parameter.
        @type param: lems.parser.expr.ExprNode
        """

        ExprNode.__init__(self, ExprNode.FUNC1)

        self.func = func
        """ Funcion stored in this node.
        @type: string """

        self.param = param
        """ Parameter.
        @type: lems.parser.expr.ExprNode """

    def __str__(self):
        """
        Generates a string representation of this node.
        """

        return '({0} {1})'.format(self.func, str(self.param))
    
    def __repr__(self):
        return self.__str__()
    
    def to_python_expr(self):
        return '({0}({1}))'.format(self.func, self.param.to_python_expr())


class ExprParser(LEMSBase):
    """
    Parser class for parsing an expression and generating a parse tree.
    """

    debug = False

    op_priority = {
        '$':-5,
        'func':8,
        '**':6,
        '+':5,
        '-':5,
        '*':6,
        '/':6,
        '^':7,
        '~':8,
        'exp':8,
        '.and.':1,
        '.or.':1,
        '.gt.':2,
        '.ge.':2,
        '.geq.':2,
        '.lt.':2,
        '.le.':2,
        '.eq.':2,
        '.neq.':2,
        '.ne.':2}  # .neq. is preferred!

    depth = 0

    """ Dictionary mapping operators to their priorities.
    @type: dict(string -> Integer) """

    def __init__(self, parse_string):
        """
        Constructor.

        @param parse_string: Expression to be parsed.
        @type parse_string: string
        """

        self.parse_string = parse_string
        """ Expression to be parsed.
        @type: string """

        self.token_list = None
        """ List of tokens from the expression to be parsed.
        @type: list(string) """

    def is_op(self, str):
        """
        Checks if a token string contains an operator.

        @param str: Token string to be checked.
        @type str: string

        @return: True if the token string contains an operator.
        @rtype: Boolean
        """

        return str in self.op_priority

    def is_func(self, str):
        """
        Checks if a token string contains a function.

        @param str: Token string to be checked.
        @type str: string

        @return: True if the token string contains a function.
        @rtype: Boolean
        """

        return str in known_functions

    def is_sym(self, str):
        """
        Checks if a token string contains a symbol.

        @param str: Token string to be checked.
        @type str: string

        @return: True if the token string contains a symbol.
        @rtype: Boolean

        MV: Added the part to recognize ** for power
        """

        return str in ['+', '-', '~', '*', '/', '^', '(', ')']

    def priority(self, op):
        if self.is_op(op):
            return self.op_priority[op]
        elif self.is_func(op):
            return self.op_priority['func']
        else:
            return self.op_priority['$']

    def tokenize(self):
        """
        Tokenizes the string stored in the parser object into a list
        of tokens.
        """

        powerflag = 0
        self.token_list = []
        ps = self.parse_string.strip()

        i = 0
        last_token = None

        while i < len(ps) and ps[i].isspace():
            i += 1

        while i < len(ps):
            token = ''

            if ps[i].isalpha():
                while i < len(ps) and (ps[i].isalnum() or ps[i] == '_'):
                    token += ps[i]
                    i += 1
            elif ps[i].isdigit():
                while i < len(ps) and (ps[i].isdigit() or
                                       ps[i] == '.' or
                                       ps[i] == 'e' or
                                       ps[i] == 'E' or
                                       (ps[i] == '+' and (ps[i-1] == 'e' or ps[i-1] == 'E')) or
                                       (ps[i] == '-' and (ps[i-1] == 'e' or ps[i-1] == 'E'))):
                    token += ps[i]
                    i += 1
            elif ps[i] == '.':
                if ps[i+1].isdigit():
                    while i < len(ps) and (ps[i].isdigit() or ps[i] == '.'):
                        token += ps[i]
                        i += 1
                else:
                    while i < len(ps) and (ps[i].isalpha() or ps[i] == '.'):
                        token += ps[i]
                        i += 1
            elif ps[i] == '*':
                while i < len(ps) and ps[i] == '*':# and j < 2:
                    token += ps[i]
                    i += 1
            else:
                token += ps[i]
                i += 1

            if token == '-' and \
               (last_token == None or last_token == '(' or self.is_op(last_token)):
                token = '~'

            self.token_list += [token]
            last_token = token

            while i < len(ps) and ps[i].isspace():
                i += 1
                
    def make_op_node(self, op, right):
        if self.is_func(op):
            return Func1Node(op, right)
        elif op == '~':
            return OpNode('-', ValueNode('0'), right)
        else:
            left = self.val_stack.pop()
            if left == '$':
                left = self.node_stack.pop()
            else:
                left = ValueNode(left)

            return OpNode(op, left, right)

    def cleanup_stacks(self):
        right = self.val_stack.pop()
        if right == '$':
            right = self.node_stack.pop()
        else:
            right = ValueNode(right)
            
        if self.debug: print('- Cleanup > right: %s'% right)
        
        while self.op_stack.top() != '$':
            if self.debug: print('5> op stack: %s, val stack: %s, node stack: %s'% ( self.op_stack, self.val_stack, self.node_stack))
            op = self.op_stack.pop()

            right = self.make_op_node(op, right)

            if self.debug: print('6> op stack: %s, val stack: %s, node stack: %s'% ( self.op_stack, self.val_stack, self.node_stack))
            if self.debug: print('7> %s'% right)
            #if self.debug: print(''

        return right

    def parse_token_list_rec(self, min_precedence):
        """
        Parses a tokenized arithmetic expression into a parse tree. It calls
        itself recursively to handle bracketed subexpressions.

        @return: Returns a token string.
        @rtype: lems.parser.expr.ExprNode

        @attention: Does not handle unary minuses at the moment. Needs to be
        fixed.
        """

        exit_loop = False

        ExprParser.depth = ExprParser.depth + 1
        if self.debug: print('>>>>> Depth: %i'% ExprParser.depth)

        precedence = min_precedence

        while self.token_list:
            token = self.token_list[0]
            la = self.token_list[1] if len(self.token_list) > 1 else None

            if self.debug: print('0> %s'% self.token_list)
            if self.debug: print('1> Token: %s, next: %s, op stack: %s, val stack: %s, node stack: %s'% (token, la, self.op_stack, self.val_stack, self.node_stack))

            self.token_list = self.token_list[1:]
            
            close_bracket = False
            
            if token == '(':
                np = ExprParser('')
                np.token_list = self.token_list

                nexp = np.parse2()

                self.node_stack.push(nexp)
                self.val_stack.push('$')

                self.token_list = np.token_list
                if self.debug: print('>>> Tokens left: %s'%self.token_list)
                close_bracket = True
            elif token == ')':
                break
            elif self.is_func(token):
                self.op_stack.push(token)
            elif self.is_op(token):
                stack_top = self.op_stack.top()
                if self.debug: print('OP Token: %s (prior: %i), top: %s (prior: %i)'% (token, self.priority(token), stack_top, self.priority(stack_top)))
                if self.priority(token) < self.priority(stack_top):
                    if self.debug: print('  Priority of %s is less than %s'%(token, stack_top))
                    self.node_stack.push(self.cleanup_stacks())
                    self.val_stack.push('$')
                else:
                    if self.debug: print('  Priority of %s is greater than %s'%(token, stack_top))
                

                self.op_stack.push(token)
            else:
                if self.debug: print('Not a bracket func or op...')
                if la == '(':
                    raise Exception("Error parsing expression: %s\nToken: %s is placed like a function but is not recognised!\nKnown functions: %s"%(self.parse_string, token, known_functions))
                stack_top = self.op_stack.top()
                if stack_top == '$':
                    if self.debug: print("option a")
                    self.node_stack.push(ValueNode(token))
                    self.val_stack.push('$')
                else:
                    if (self.is_op(la) and
                        self.priority(stack_top) < self.priority(la)):
                        if self.debug: print("option b")

                        self.node_stack.push(ValueNode(token))
                        self.val_stack.push('$')
                    else:
                        if self.debug: print("option c, nodes: %s"% self.node_stack)
                        op = self.op_stack.pop()

                        right = ValueNode(token)
                        op_node = self.make_op_node(op,right)

                        self.node_stack.push(op_node)
                        self.val_stack.push('$')
                        
            if close_bracket:
                stack_top = self.op_stack.top()
                if self.debug: print("+ Closing bracket, op stack: %s, node stack: %s la: %s"%(self.op_stack, self.node_stack, la))
                if self.debug: print('>>> Tokens left: %s'%self.token_list)
                
                if stack_top == '$':
                    if self.debug: print("+ option a")
                    '''
                    self.node_stack.push(ValueNode(token))
                    self.val_stack.push('$')'''
                else:
                    la = self.token_list[0] if len(self.token_list) > 1 else None
                    if (self.is_op(la) and self.priority(stack_top) < self.priority(la)):
                        if self.debug: print("+ option b")
                        #self.node_stack.push(ValueNode(token))
                        #self.val_stack.push('$')
                    else:
                        if self.debug: print("+ option c, nodes: %s"% self.node_stack)
                        if self.debug: print('35> op stack: %s, val stack: %s, node stack: %s'% ( self.op_stack, self.val_stack, self.node_stack))
                        right = self.node_stack.pop()
                        op = self.op_stack.pop()
                        op_node = self.make_op_node(stack_top,right)
                        if self.debug: print("Made op node: %s, right: %s"%(op_node, right))

                        self.node_stack.push(op_node)
                        self.val_stack.push('$')
                        if self.debug: print('36> op stack: %s, val stack: %s, node stack: %s'% ( self.op_stack, self.val_stack, self.node_stack))
                        
            

            if self.debug: print('2> Token: %s, next: %s, op stack: %s, val stack: %s, node stack: %s'% (token, la, self.op_stack, self.val_stack, self.node_stack))
            if self.debug: print('')

        if self.debug: print('3> op stack: %s, val stack: %s, node stack: %s'% ( self.op_stack, self.val_stack, self.node_stack))
        ret = self.cleanup_stacks()

        if self.debug: print('4> op stack: %s, val stack: %s, node stack: %s'% ( self.op_stack, self.val_stack, self.node_stack))
        if self.debug: print('<<<<< Depth: %s, returning: %s'% (ExprParser.depth, ret))
        ExprParser.depth = ExprParser.depth - 1
        if self.debug: print('')
        return ret

    def parse(self):
        """
        Tokenizes and parses an arithmetic expression into a parse tree.

        @return: Returns a token string.
        @rtype: lems.parser.expr.ExprNode
        """
        #print("Parsing: %s"%self.parse_string)
        self.tokenize()
        if self.debug: print("Tokens found: %s"%self.token_list)

        try:
            parse_tree = self.parse2()
        except Exception as e:
                raise e
        return parse_tree

    def parse2(self):
        self.op_stack = Stack()
        self.val_stack = Stack()
        self.node_stack = Stack()

        self.op_stack.push('$')
        self.val_stack.push('$')
        try:
            ret = self.parse_token_list_rec(self.op_priority['$'])
        except Exception as e:
            raise e
    
        return ret

    def __str__(self):
        return str(self.token_list)



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#


class TraitError(Exception):
    def __init__(self, msg='', trait=None, attr=None):
        self.trait = trait
        self.attr = attr
        self.message = msg
        super(TraitError, self).__init__(msg)

    def __str__(self):
        lines = [self.args[0]]
        if self.attr:
            lines.append('  attribute {}'.format(self.attr))
        if self.trait:
            lines.append('  class {}'.format(type(self.trait).__name__))

        return '\n'.join(lines)


class TraitAttributeError(TraitError, AttributeError):
    pass


class TraitFinalAttributeError(TraitAttributeError, AttributeError):
    pass


class TraitValueError(TraitError, ValueError):
    pass


class TraitTypeError(TraitError, TypeError):
    pass



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

import numpy as np
import scipy.sparse as ss

from .backendtestbase import BaseTestSim
from tvb.simulator.backend.nb_mpr import NbMPRBackend
from tvb.simulator.lab import *


class TestNbSim(BaseTestSim):

    def _get_run_sim(self, print_source=False):
        template = '<%include file="nb-montbrio.py.mako"/>'
        content = dict(foo='bar')

        return NbMPRBackend().build_py_func(template, content, name='run_sim', print_source=print_source)

    def _random_network(self, N=500, density=0.1, low=5, high=250, speed=np.inf):
        weights = ss.random(N,N, density=density, format='lil')
        weights.setdiag(0)
        weights = weights.tocsr()
        lengths = weights.copy()
        lengths.data[:] = np.random.uniform(low=low, high=high, size=lengths.data.shape[0])

        N = weights.shape[0]
        conn = connectivity.Connectivity(
                weights=weights.A,
                tract_lengths=lengths.A,
                region_labels=np.array( [f'roi_{i}' for i in range(N)]), 
                centres=np.zeros(N),
                speed=np.r_[speed]
        )
        return conn

    def test_import(self):
        run_sim = self._get_run_sim(print_source=True)

    def test_local_deterministic(self):
        run_sim = self._get_run_sim(print_source=True)
        dt = 0.01
        G = 0.

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Scaling(a=np.array([G])),
            connectivity=self._random_network(),
            conduction_speed=np.inf,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunStochastic( # matching integrator
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.0, 0.0]),
                    noise_seed=42
                )
            )
        ).configure()

        r_pdq, V_pdq = run_sim(sim, nstep=1)

        # check initial conditions
        np.testing.assert_allclose(sim.current_state[0,:,0], r_pdq[:,0])
        np.testing.assert_allclose(sim.current_state[1,:,0], V_pdq[:,0])

        (raw_t, raw_d), = sim.run(simulation_length=1)
        r_tvb, V_tvb = raw_d[0,:,:,0]
        r_pdq = r_pdq[:,1] # we include initial state, TVB doesn't
        V_pdq = V_pdq[:,1]

        np.testing.assert_allclose(r_tvb, r_pdq, rtol=1e-5)
        np.testing.assert_allclose(V_tvb, V_pdq, rtol=1e-5)


    def test_local_stochastic(self):
        run_sim = self._get_run_sim(print_source=True)
        dt = 0.01
        G = 0.

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Scaling(a=np.array([G])),
            connectivity=self._random_network(),
            conduction_speed=np.inf,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunStochastic( 
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.01, 0.02]),
                    noise_seed=42
                )
            )
        ).configure()

        sim_det = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Scaling(a=np.array([G])),
            connectivity=self._random_network(),
            conduction_speed=np.inf,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunDeterministic( 
                dt=dt, 
            )
        ).configure()

        r_pdq, V_pdq = run_sim(sim, nstep=200)

        # check initial conditions
        np.testing.assert_allclose(sim.current_state[0,:,0], r_pdq[:,0])
        np.testing.assert_allclose(sim.current_state[1,:,0], V_pdq[:,0])

        (raw_t, raw_d), = sim.run(simulation_length=1)
        (raw_t_det, raw_d_det), = sim_det.run(simulation_length=1)
        r_tvb, V_tvb = raw_d[0,:,:,0]
        r_tvb_det, V_tvb_det = raw_d_det[0,:,:,0]
        r_pdq = r_pdq[:,1] # we include initial state, TVB doesn't
        V_pdq = V_pdq[:,1]

        np.testing.assert_allclose(
                np.mean(r_tvb_det - r_tvb),
                np.mean(r_tvb_det - r_pdq),
                atol=1e-2 # zero mean doesn't play well with rtol
        )
        np.testing.assert_allclose(
                np.mean(V_tvb_det - V_tvb),
                np.mean(V_tvb_det - V_pdq),
                atol=1e-2
        )
        np.testing.assert_allclose(
                np.std(r_tvb_det - r_tvb), 
                np.std(r_tvb_det - r_pdq),
                rtol=1e-2
        )
        np.testing.assert_allclose(
                np.std(V_tvb_det - V_tvb),
                np.std(V_tvb_det - V_pdq),
                rtol=1e-2
        )

    def test_network_deterministic_nodelay(self):
        run_sim = self._get_run_sim(print_source=True)
        dt = 0.01
        G = 0.8

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Scaling(a=np.array([G])),
            connectivity=self._random_network(),
            conduction_speed=np.inf,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunStochastic( 
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.0, 0.0]),
                    noise_seed=42
                )
            )
        ).configure()

        r_pdq, V_pdq = run_sim(sim, nstep=1)

        np.testing.assert_allclose(sim.current_state[0,:,0], r_pdq[:,0])
        np.testing.assert_allclose(sim.current_state[1,:,0], V_pdq[:,0])
        r_pdq = r_pdq[:,1] # we include initial state, TVB doesn't
        V_pdq = V_pdq[:,1]

        (raw_t, raw_d), = sim.run(simulation_length=1)
        r_tvb, V_tvb = raw_d[0,:,:,0]

        np.testing.assert_allclose(r_tvb, r_pdq, rtol=1e-4)
        np.testing.assert_allclose(V_tvb, V_pdq, rtol=1e-4)
        
    def test_network_deterministic_delay(self):
        run_sim = self._get_run_sim(print_source=True)
        dt = 0.01
        G = 0.
        speed=2.

        sim = simulator.Simulator(
            model=models.MontbrioPazoRoxin(),
            coupling=coupling.Scaling(a=np.array([G])),
            connectivity=self._random_network(speed=speed),
            conduction_speed=speed,
            monitors=[
                monitors.Raw()
            ],
            integrator=integrators.HeunStochastic( 
                dt=dt, 
                noise=noise.Additive(
                    nsig=np.array([0.0, 0.0]),
                    noise_seed=42
                )
            )
        ).configure()

        r_pdq, V_pdq = run_sim(sim, nstep=1)

        np.testing.assert_allclose(sim.current_state[0,:,0], r_pdq[:,sim.connectivity.horizon-1])
        np.testing.assert_allclose(sim.current_state[1,:,0], V_pdq[:,sim.connectivity.horizon-1])
        r_pdq = r_pdq[:,-1] # we include initial state, TVB doesn't
        V_pdq = V_pdq[:,-1]

        (raw_t, raw_d), = sim.run(simulation_length=1)
        r_tvb, V_tvb = raw_d[0,:,:,0]

        np.testing.assert_allclose(r_tvb, r_pdq, rtol=1e-4)
        np.testing.assert_allclose(V_tvb, V_pdq, rtol=1e-4)



from mako.template import Template

import os
import sys

# not ideal but avoids modifying  the vendored LEMS itself
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from lems.model.model import Model


def default_lems_folder():
    here = os.path.dirname(os.path.abspath(__file__))
    print('\n here', here, '\n')
    xmlpath = os.path.join(here, 'XMLmodels')
    return xmlpath

def lems_file(model_name, folder=None):
    folder = folder or default_lems_folder()
    return os.path.join(folder, model_name.lower() + '_CUDA.xml')

def default_template():
    here = os.path.dirname(os.path.abspath(__file__))
    tmp_filename = os.path.join(here, 'tmpl8_CUDA.py')
    template = Template(filename=tmp_filename)
    return template

def load_model(model_filename, folder=None):
    "Load model from filename"

    fp_xml = lems_file(model_filename, folder)

    model = Model()
    model.import_from_file(fp_xml)
    # modelextended = model.resolve()

    return model

def render_model(model_name, template=None, folder=None):

    model = load_model(model_name, folder)
    template = template or default_template()

    modellist = model.component_types[model_name]

    # coupling functionality
    couplinglist = list()

    for i, cplists in enumerate(model.component_types):
        if 'coupling' in cplists.name:
            couplinglist.append(cplists)

    # collect total number of exposures combinations.
    expolist = list()
    for i, expo in enumerate(modellist.exposures):
        for chc in expo.choices:
            expolist.append(chc)

    # only check whether noise is there, if so then activate it
    noisepresent=False
    for ct in (model.component_types):
        if ct.name == 'noise' and ct.description == 'on':
            noisepresent=True

    # start templating
    model_str = template.render(
                            modelname=model_name,
                            const=modellist.constants,
                            dynamics=modellist.dynamics,
                            params=modellist.parameters,
                            derparams=modellist.derived_parameters,
                            coupling=couplinglist,
                            noisepresent=noisepresent,
                            expolist=expolist
                            )

    return model_str

def cuda_templating(model_filename, folder=None):

    modelfile = os.path.join((os.path.dirname(os.path.abspath(__file__))), 'CUDAmodels', model_filename.lower() + '.c')

    # start templating
    model_str = render_model(model_filename, template=default_template(), folder=folder)

    # write template to file
    with open(modelfile, "w") as f:
        f.writelines(model_str)

if __name__ == '__main__':

    # model_filename = 'Oscillator'
    # model_filename = 'Kuramoto'
    # model_filename = 'Rwongwang'
    model_filename = 'Epileptor'
    cuda_templating(model_filename)



// -*- coding: utf-8 -*-
//
//
// TheVirtualBrain-Scientific Package. This package holds all simulators, and
// analysers necessary to run brain-simulations. You can use it stand alone or
// in conjunction with TheVirtualBrain-Framework Package. See content of the
// documentation-folder for more details. See also http://www.thevirtualbrain.org
//
// (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
//
// This program is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
// PARTICULAR PURPOSE.  See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with this
// program.  If not, see <http://www.gnu.org/licenses/>.
//
//
//   CITATION:
// When using The Virtual Brain for scientific publications, please cite it as follows:
//
//  Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
//  Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
//      The Virtual Brain: a simulator of primate brain network dynamics.
//   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
//
//

#include <stdio.h> // for printf
#define PI_2 (2 * M_PI_F)

// buffer length defaults to the argument to the integrate kernel
// but if it's known at compile time, it can be provided which allows
// compiler to change i%n to i&(n-1) if n is a power of two.
#ifndef NH
#define NH nh
#endif

#ifndef WARP_SIZE
#define WARP_SIZE 32
#endif

#include <curand_kernel.h>
#include <curand.h>
#include <stdbool.h>

__device__ float wrap_it_PI(float x)
{
    bool neg_mask = x < 0.0f;
    bool pos_mask = !neg_mask;
    // fmodf diverges 51% of time
    float pos_val = fmodf(x, PI_2);
    float neg_val = PI_2 - fmodf(-x, PI_2);
    return neg_mask * neg_val + pos_mask * pos_val;
}
\
% for state_var in (dynamics.state_variables):
% if (state_var.boundaries != "PI"):
__device__ float wrap_it_${state_var.name}(float ${state_var.name})
{
    float ${state_var.name}dim[] = {${state_var.boundaries}};
    if (${state_var.name} < ${state_var.name}dim[0]) ${state_var.name} = ${state_var.name}dim[0];
    else if (${state_var.name} > ${state_var.name}dim[1]) ${state_var.name} = ${state_var.name}dim[1];

    return ${state_var.name};
}
% endif /
% endfor

__global__ void ${modelname}(

        // config
        unsigned int i_step, unsigned int n_node, unsigned int nh, unsigned int n_step, unsigned int n_params,
        float dt, float speed, float * __restrict__ weights, float * __restrict__ lengths,
        float * __restrict__ params_pwi, // pwi: per work item
        // state
        float * __restrict__ state_pwi,
        // outputs
        float * __restrict__ tavg_pwi
        )
{
    // work id & size
    const unsigned int id = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x + threadIdx.x;
    const unsigned int size = blockDim.x * gridDim.x * gridDim.y;

#define params(i_par) (params_pwi[(size * (i_par)) + id])
#define state(time, i_node) (state_pwi[((time) * ${dynamics.state_variables.__len__()} * n_node + (i_node))*size + id])
#define tavg(i_node) (tavg_pwi[((i_node) * size) + id])

    // unpack params
    // These are the two parameters which are usually explore in fitting in this model
    ## printing the to be sweeped parameters
    % for paramcounter, par_var in enumerate(params):
    const ${par_var.dimension} ${par_var.name} = params(${paramcounter});
    % endfor

    // regular constants
% for item in const:
    const float ${item.name} = ${item.default};
% endfor /

    // coupling constants, coupling itself is hardcoded in kernel
% for m in range(len(coupling)):
    % for cc in (coupling[m].constants):
    const float ${cc.name} = ${cc.default};
    %endfor /
% endfor

    // coupling parameters
% for m in range(len(coupling)):
    % for cc in (coupling[m].derived_parameters):
    float ${cc.name} = 0.0;
    %endfor /
% endfor

    % if derparams:
    // derived parameters
    % for par_var in derparams:
    const float ${par_var.name} = ${par_var.expression};
    % endfor /
    %endif /

    % if dynamics.derived_variables:
    // the dynamic derived variables declarations
    % for i, dv in enumerate(dynamics.derived_variables):
    float ${dv.name} = 0.0;
    % endfor /
    % endif /

    % if dynamics.conditional_derived_variables:
    // conditional_derived variable declaration
    % for cd in dynamics.conditional_derived_variables:
    float ${cd.name} = 0.0;
    % endfor /
    % endif /

    curandState crndst;
    curand_init(id * (blockDim.x * gridDim.x * gridDim.y), 0, 0, &crndst);

    % for state_var in (dynamics.state_variables):
    float ${state_var.name} = 0.0;
    % endfor

    % for td in (dynamics.time_derivatives):
    float ${td.name} = 0.0;
    % endfor /

    //***// This is only initialization of the observable
    for (unsigned int i_node = 0; i_node < n_node; i_node++)
    {
        tavg(i_node) = 0.0f;
        if (i_step == 0){
            state(i_step, i_node) = 0.001;
        }
    }

    //***// This is the loop over time, should stay always the same
    for (unsigned int t = i_step; t < (i_step + n_step); t++)
    {
    //***// This is the loop over nodes, which also should stay the same
        for (int i_node = 0; i_node < n_node; i_node++)
        {
            % for m in range(len(coupling)):
                % for cdp in (coupling[m].derived_parameters):
            ${cdp.name} = 0.0f;
                %endfor
            % endfor /
            ## float coupling = 0.0f;

            % for i, item in enumerate(dynamics.state_variables):
            ${item.name} = state((t) % nh, i_node + ${i} * n_node);
            % endfor /

            // This variable is used to traverse the weights and lengths matrix, which is really just a vector. It is just a displacement. /
            unsigned int i_n = i_node * n_node;

            for (unsigned int j_node = 0; j_node < n_node; j_node++)
            {
                //***// Get the weight of the coupling between node i and node j
                float wij = weights[i_n + j_node]; // nb. not coalesced
                if (wij == 0.0)
                    continue;

                % if (derparams['rec_speed_dt'] and derparams['rec_speed_dt'].expression != '0'):
                //***// Get the delay between node i and node j
                unsigned int dij = lengths[i_n + j_node] * rec_speed_dt;
                % else:
                // no delay specified
                unsigned int dij = 0;
                % endif

                //***// Get the state of node j which is delayed by dij
                % for m in range(len(coupling)):
                    % for cp in (coupling[m].parameters):
                float ${cp.name} = state(((t - dij + nh) % nh), j_node + ${cp.dimension} * n_node);
                    % endfor /
                %endfor

                // Sum it all together using the coupling function. Kuramoto coupling: (postsyn * presyn) == ((a) * (sin(xj - xi))) \
                % for ml in range(len(coupling)):
                    ## only do this if pre or post is specified
                    % if coupling[ml].dynamics.derived_variables and \
                        (coupling[ml].dynamics.derived_variables['pre'].expression != 'None' or \
                         coupling[ml].dynamics.derived_variables['post'].expression != 'None'):
                    % for cdp in (coupling[ml].derived_parameters):

                ${cdp.name} += wij * ${coupling[ml].dynamics.derived_variables['post'].expression} * ${coupling[ml].dynamics.derived_variables['pre'].expression};
                ## coupling += wij * ${coupling[ml].dynamics.derived_variables['post'].expression} * ${coupling[ml].dynamics.derived_variables['pre'].expression};

                    %endfor
                    % endif /
                % endfor /
            } // j_node */

            // rec_n is used for the scaling over nodes
            % for m in range(len(coupling)):
                % for cdp in (coupling[m].derived_parameters):
                    % if cdp.expression and (cdp.expression !='None' and cdp.expression !='none'):
            ${cdp.name} *= ${cdp.expression};
                    % endif /
                % endfor
            % endfor \

            % if dynamics.derived_variables:
            // the dynamic derived variables
            % for i, dv in enumerate(dynamics.derived_variables):
            ${dv.name} = ${dv.expression};
            % endfor /
            % endif

            % for con_der in dynamics.conditional_derived_variables:
            if (${con_der.condition})
                // The conditional variables
                % for case in (con_der.cases):
                    % if (loop.first):
                ${con_der.name} = ${case};
                    % elif (loop.last and not loop.first):
            else
                ${con_der.name} = ${case};
                    %endif /
                % endfor
            % endfor \

            // This is dynamics step and the update in the state of the node
            % for i, tim_der in enumerate(dynamics.time_derivatives):
            ${tim_der.name} = dt * (${tim_der.expression});
            % endfor

            % if noisepresent:
            // Add noise (if noise components are present in model), integrate with stochastic forward euler and wrap it up
            % for ds, td in zip(dynamics.state_variables, dynamics.time_derivatives):
            ${ds.name} += nsig * curand_normal(&crndst) + ${td.name};
            % endfor /
            ##% else:
            ##% for ds, td in zip(dynamics.state_variables, dynamics.time_derivatives):
            ##${ds.name} += dt a * ${td.name});
            ##% endfor /
            % endif

            // Wrap it within the limits of the model
            % for state_var in (dynamics.state_variables):
                % if state_var.boundaries == 'PI':
            ${state_var.name} = wrap_it_${state_var.boundaries}(${state_var.name});
                % else:
            ${state_var.name} = wrap_it_${state_var.name}(${state_var.name});
                % endif
            % endfor /

            // Update the state
            % for i, state_var in enumerate(dynamics.state_variables):
            state((t + 1) % nh, i_node + ${i} * n_node) = ${state_var.name};
            % endfor /

            // Update the observable only for the last timestep
            if (t == (i_step + n_step - 1)){
                % for i, expo in enumerate(expolist):
                tavg(i_node + ${i} * n_node) = ${expo};
                % endfor /
            }

            // sync across warps executing nodes for single sim, before going on to next time step
            __syncthreads();

        } // for i_node
    } // for t

// cleanup macros/*{{{*/
#undef params
#undef state
#undef tavg/*}}}*/

} // kernel integrate


"""
LEMS XML file format parser.

@author: Gautham Ganapathy
@organization: LEMS (http://neuroml.org/lems/, https://github.com/organizations/LEMS)
@contact: gautham@lisphacker.org

MAvdVlag: altered attributes for constants, state_variables, derived_variables, time_derivatives,
conditional_derived_variable and exposures
"""

import xml.etree.ElementTree as xe

from lems.base.base import LEMSBase
# from lems.model.fundamental import *
from lems.model.component import *
# from lems.model.dynamics import *
# from lems.model.structure import *
# from lems.model.simulation import *

# from lems.base.util import make_id

from lems.model.dynamics import *


def get_nons_tag_from_node(node):
    tag = node.tag
    bits = tag.split('}')
    if len(bits) == 1:
        return tag
    elif '/lems/' in bits[0]:
        return bits[1]
    elif 'neuroml2' in bits[0]:
        return bits[1]
    elif 'rdf' in bits[0]:
        return "rdf_"+bits[1]
    elif 'model-qualifiers' in bits[0]:
        return "bqmodel_"+bits[1]
    elif 'biology-qualifiers' in bits[0]:
        return "bqbiol_"+bits[1]
    else:
        return "%s:%s"%(bits[0],bits[1])

class LEMSXMLNode:
    def __init__(self, pyxmlnode):
        self.tag = get_nons_tag_from_node(pyxmlnode)
        self.ltag = self.tag.lower()

        self.attrib = dict()
        self.lattrib = dict()

        for k in pyxmlnode.attrib:
            self.attrib[k] = pyxmlnode.attrib[k]
            self.lattrib[k.lower()] = pyxmlnode.attrib[k]

        self.children = list()
        for pyxmlchild in pyxmlnode:
            self.children.append(LEMSXMLNode(pyxmlchild))
            
    def __str__(self):
        return 'LEMSXMLNode <{0} {1}>'.format(self.tag, self.attrib)
        
class LEMSFileParser(LEMSBase):
    """
    LEMS XML file format parser class.
    """
    
    def __init__(self, model, include_dirs = [], include_includes=True):
        """
        Constructor.

        See instance variable documentation for more details on parameters.
        """

        self.model = model
        """ Model instance to be populated from the parsed file.
        @type: lems.model.model.Model """

        self.include_dirs = include_dirs
        """ List of directories to search for included files.
        @type: list(str) """

        self.tag_parse_table = None
        """ Dictionary of xml tags to parse methods
        @type: dict(string -> function) """

        self.valid_children = None
        """ Dictionary mapping each tag to it's list of valid child tags.
        @type: dict(string -> string) """

        self.id_counter = None
        """ Counter generator for generating unique ids.
        @type: generator(int) """
        
        self.include_includes = include_includes
        """ Whether to include LEMS definitions in <Include> elements
        @type: boolean """
        
        self.init_parser()

    def init_parser(self):
        """
        Initializes the parser
        """

        #self.token_list = None
        #self.prev_token_lists = None

        self.valid_children = dict()
        self.valid_children['lems'] = ['component', 'componenttype',
                                       'target', 'include',
                                       'dimension', 'unit', 'assertion']
                                       
        #TODO: make this generic for any domain specific language based on LEMS
        self.valid_children['neuroml'] = ['include', 'componenttype']
                                       
        self.valid_children['componenttype'] = ['dynamics',
                                                'child', 'children',
                                                'componentreference',
                                                'exposure', 'eventport',
                                                'fixed', 'link', 'parameter',
                                                'property',
                                                'indexparameter',
                                                'path', 'requirement',
                                                'componentrequirement',
                                                'instancerequirement',
                                                'simulation', 'structure',
                                                'text', 'attachments',
                                                'constant', 'derivedparameter',
                                                'function']
                                                
        self.valid_children['dynamics'] = ['derivedvariable',
                                           'conditionalderivedvariable',
                                           'oncondition',
                                           'onevent', 'onstart',
                                           'statevariable', 'timederivative',
                                           'kineticscheme', 'regime']
                                           
        self.valid_children['component'] = ['component']
                                           
        self.valid_children['conditionalderivedvariable'] = ['case']
        
        self.valid_children['regime'] = ['oncondition', 'onentry', 'timederivative']
        self.valid_children['oncondition'] = ['eventout', 'stateassignment', 'transition']
        self.valid_children['onentry'] = ['eventout', 'stateassignment', 'transition']
        self.valid_children['onevent'] = ['eventout', 'stateassignment', 'transition']
        self.valid_children['onstart'] = ['eventout', 'stateassignment', 'transition']
        self.valid_children['structure'] = ['childinstance',
                                            'eventconnection',
                                            'foreach',
                                            'multiinstantiate',
                                            'with',
                                            'tunnel']
                                       
        self.valid_children['foreach'] = ['foreach', 'eventconnection']     
                                            
        self.valid_children['simulation'] = ['record', 'eventrecord', 'run',
                                             'datadisplay', 'datawriter', 'eventwriter']

        self.tag_parse_table = dict()
        #self.tag_parse_table['assertion'] = self.parse_assertion
        self.tag_parse_table['attachments'] = self.parse_attachments
        self.tag_parse_table['child'] = self.parse_child
        self.tag_parse_table['childinstance'] = self.parse_child_instance
        self.tag_parse_table['children'] = self.parse_children
        self.tag_parse_table['component'] = self.parse_component
        self.tag_parse_table['componentreference'] = self.parse_component_reference   
        self.tag_parse_table['componentrequirement'] = self.parse_component_requirement
        self.tag_parse_table['componenttype'] = self.parse_component_type
        self.tag_parse_table['constant'] = self.parse_constant
        self.tag_parse_table['function'] = self.parse_function
        self.tag_parse_table['datadisplay'] = self.parse_data_display
        self.tag_parse_table['datawriter'] = self.parse_data_writer
        self.tag_parse_table['eventwriter'] = self.parse_event_writer
        self.tag_parse_table['derivedparameter'] = self.parse_derived_parameter
        self.tag_parse_table['derivedvariable'] = self.parse_derived_variable
        self.tag_parse_table['conditionalderivedvariable'] = self.parse_conditional_derived_variable
        self.tag_parse_table['case'] = self.parse_case
        self.tag_parse_table['dimension'] = self.parse_dimension
        self.tag_parse_table['dynamics'] = self.parse_dynamics
        self.tag_parse_table['eventconnection'] = self.parse_event_connection
        self.tag_parse_table['eventout'] = self.parse_event_out
        self.tag_parse_table['eventport'] = self.parse_event_port
        self.tag_parse_table['exposure'] = self.parse_exposure
        self.tag_parse_table['fixed'] = self.parse_fixed
        self.tag_parse_table['foreach'] = self.parse_for_each
        self.tag_parse_table['include'] = self.parse_include
        self.tag_parse_table['indexparameter'] = self.parse_index_parameter
        self.tag_parse_table['kineticscheme'] = self.parse_kinetic_scheme
        self.tag_parse_table['link'] = self.parse_link
        self.tag_parse_table['multiinstantiate'] = self.parse_multi_instantiate
        self.tag_parse_table['oncondition'] = self.parse_on_condition
        self.tag_parse_table['onentry'] = self.parse_on_entry
        self.tag_parse_table['onevent'] = self.parse_on_event
        self.tag_parse_table['onstart'] = self.parse_on_start
        self.tag_parse_table['parameter'] = self.parse_parameter
        self.tag_parse_table['property'] = self.parse_property
        self.tag_parse_table['path'] = self.parse_path
        self.tag_parse_table['record'] = self.parse_record
        self.tag_parse_table['eventrecord'] = self.parse_event_record
        self.tag_parse_table['regime'] = self.parse_regime
        self.tag_parse_table['requirement'] = self.parse_requirement
        self.tag_parse_table['instancerequirement'] = self.parse_instance_requirement     
        self.tag_parse_table['run'] = self.parse_run
        #self.tag_parse_table['show'] = self.parse_show
        self.tag_parse_table['simulation'] = self.parse_simulation
        self.tag_parse_table['stateassignment'] = self.parse_state_assignment
        self.tag_parse_table['statevariable'] = self.parse_state_variable
        self.tag_parse_table['structure'] = self.parse_structure
        self.tag_parse_table['target'] = self.parse_target
        self.tag_parse_table['text'] = self.parse_text
        self.tag_parse_table['timederivative'] = self.parse_time_derivative
        self.tag_parse_table['transition'] = self.parse_transition
        self.tag_parse_table['tunnel'] = self.parse_tunnel
        self.tag_parse_table['unit'] = self.parse_unit
        self.tag_parse_table['with'] = self.parse_with

        self.xml_node_stack = []

        self.current_component_type = None
        self.current_dynamics = None
        self.current_regime = None
        self.current_event_handler = None
        self.current_structure = None
        self.current_simulation = None
        self.current_component = None
        
        def counter():
            count = 1
            while True:
                yield count
                count = count + 1

        self.id_counter = counter()

        
    def process_nested_tags(self, node, tag = ''):
        """
        Process child tags.

        @param node: Current node being parsed.
        @type node: xml.etree.Element

        @raise ParseError: Raised when an unexpected nested tag is found.
        """
        ##print("---------Processing: %s, %s"%(node.tag,tag))

        if tag == '':
            t = node.ltag
        else:
            t = tag.lower()
        
        for child in node.children:
            self.xml_node_stack = [child] + self.xml_node_stack

            ctagl = child.ltag

            if ctagl in self.tag_parse_table and ctagl in self.valid_children[t]:
                #print("Processing known type: %s"%ctagl)
                self.tag_parse_table[ctagl](child)
            else:
                #print("Processing unknown type: %s"%ctagl)
                self.parse_component_by_typename(child, child.tag)

            self.xml_node_stack = self.xml_node_stack[1:]

    def parse(self, xmltext):
        """
        Parse a string containing LEMS XML text.

        @param xmltext: String containing LEMS XML formatted text.
        @type xmltext: str
        """
        
        xml = LEMSXMLNode(xe.XML(xmltext))

        if xml.ltag != 'lems' and xml.ltag != 'neuroml':
            raise ParseError('<Lems> expected as root element (or even <neuroml>), found: {0}'.format(xml.ltag))
        '''
        if xml.ltag == 'lems':
            if 'description' in xml.lattrib:
                self.description = xml.lattrib['description']
        '''

        self.process_nested_tags(xml)


    def raise_error(self, message, *params, **key_params):
        """
        Raise a parse error.
        """
        
        s = 'Parser error in '

        self.xml_node_stack.reverse()
        if len(self.xml_node_stack) > 1:
            node = self.xml_node_stack[0]
            s += '<{0}'.format(node.tag)
            if 'name' in node.lattrib:
                s += ' name=\"{0}\"'.format(node.lattrib['name'])
            if 'id' in node.lattrib:
                s += ' id=\"{0}\"'.format(node.lattrib['id'])
            s += '>'

        for node in self.xml_node_stack[1:]:
            s += '.<{0}'.format(node.tag)
            if 'name' in node.lattrib:
                s += ' name=\"{0}\"'.format(node.lattrib['name'])
            if 'id' in node.lattrib:
                s += ' id=\"{0}\"'.format(node.lattrib['id'])
            s += '>'

        s += ':\n  ' + message

        raise ParseError(s, *params, **key_params)

        self.xml_node_stack.reverse()


#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################





    def parse_assertion(self, node):
        """
        Parses <Assertion>

        @param node: Node containing the <Assertion> element
        @type node: xml.etree.Element
        """

        print('TODO - <Assertion>')


    def parse_attachments(self, node):
        """
        Parses <Attachments>

        @param node: Node containing the <Attachments> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<Attachments> must specify a name.')

        if 'type' in node.lattrib:
            type_ = node.lattrib['type']
        else:
            self.raise_error("Attachment '{0}' must specify a type.",
                             name)

        description = node.lattrib.get('description', '')
        self.current_component_type.add_attachments(Attachments(name, type_, description))

    def parse_child(self, node):
        """
        Parses <Child>

        @param node: Node containing the <Child> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<Child> must specify a name.')

        if 'type' in node.lattrib:
            type_ = node.lattrib['type']
        else:
            self.raise_error("Child '{0}' must specify a type.", name)

        self.current_component_type.add_children(Children(name, type_, False))

    def parse_child_instance(self, node):
        """
        Parses <ChildInstance>

        @param node: Node containing the <ChildInstance> element
        @type node: xml.etree.Element
        """

        if 'component' in node.lattrib:
            component = node.lattrib['component']
        else:
            self.raise_error('<ChildInstance> must specify a component reference')

        self.current_structure.add_child_instance(ChildInstance(component))

    def parse_children(self, node):
        """
        Parses <Children>

        @param node: Node containing the <Children> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<Children> must specify a name.')

        if 'type' in node.lattrib:
            type_ = node.lattrib['type']
        else:
            self.raise_error("Children '{0}' must specify a type.", name)

        self.current_component_type.add_children(Children(name, type_, True))

    def parse_component_by_typename(self, node, type_):
        """
        Parses components defined directly by component name.

        @param node: Node containing the <Component> element
        @type node: xml.etree.Element

        @param type_: Type of this component.
        @type type_: string

        @raise ParseError: Raised when the component does not have an id.
        """
        #print('Parsing component {0} by typename {1}'.format(node, type_))
        if 'id' in node.lattrib:
            id_ = node.lattrib['id']
        else:
            #self.raise_error('Component must have an id')
            id_ = node.tag #make_id()

        if 'type' in node.lattrib:
            type_ = node.lattrib['type']
        else:
            type_ = node.tag

        component = Component(id_, type_)

        if self.current_component:
            component.set_parent_id(self.current_component.id)
            self.current_component.add_child(component)
            
        else:
            self.model.add_component(component)

        for key in node.attrib:
            if key.lower() not in ['id', 'type']:
                component.set_parameter(key, node.attrib[key])

        old_component = self.current_component
        self.current_component = component
        self.process_nested_tags(node, 'component')
        self.current_component = old_component

    def parse_component(self, node):
        """
        Parses <Component>

        @param node: Node containing the <Component> element
        @type node: xml.etree.Element
        """

        if 'id' in node.lattrib:
            id_ = node.lattrib['id']
        else:
            #self.raise_error('Component must have an id')
            id_ = make_id()

        if 'type' in node.lattrib:
            type_ = node.lattrib['type']
        else:
                self.raise_error("Component {0} must have a type.",
                                 id_)

        component = Component(id_, type_)

        if self.current_component:
            component.set_parent_id(self.current_component.id)
            self.current_component.add_child(component)
        else:
            self.model.add_component(component)
                
        for key in node.attrib:
            if key.lower() not in ['id', 'type']:
                component.set_parameter(key, node.attrib[key])

        old_component = self.current_component
        self.current_component = component
        self.process_nested_tags(node)
        self.current_component = old_component

    def parse_component_reference(self, node):
        """
        Parses <ComponentReference>

        @param node: Node containing the <ComponentTypeRef> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<ComponentReference> must specify a name for the ' +
                             'reference.')

        if 'type' in node.lattrib:
            type_ = node.lattrib['type']
        else:
            self.raise_error('<ComponentReference> must specify a type for the ' +
                             'reference.')
                             
        if 'local' in node.lattrib:
            local = node.lattrib['local']
        else:
            local = None

        self.current_component_type.add_component_reference(ComponentReference(name, type_, local))

    def parse_component_type(self, node):
        """
        Parses <ComponentType>

        @param node: Node containing the <ComponentType> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when the component type does not have a
        name.
        """

        try:
            name = node.lattrib['name']
        except:
            self.raise_error('<ComponentType> must specify a name')

        if 'extends' in node.lattrib:
            extends = node.lattrib['extends']
        else:
            extends = None

        if 'description' in node.lattrib:
            description = node.lattrib['description']
        else:
            description = ''

        component_type = ComponentType(name, description, extends)
        self.model.add_component_type(component_type)

        self.current_component_type = component_type
        self.process_nested_tags(node)
        self.current_component_type = None

    def parse_constant(self, node):
        """
        Parses <Constant>

        @param node: Node containing the <Constant> element
        @type node: xml.etree.Element

        MV: fixed the symbol part. It was not there for constant parsing
        """

        try:
            name = node.lattrib['name']
        except:
            self.raise_error('<Constant> must specify a name.')

        domain = node.lattrib.get('domain', None)
        symbol = node.lattrib.get('symbol', None)

        try:
            default = node.lattrib['default']
        except:
            self.raise_error("Constant '{0}' must have a value.", name)

        description = node.lattrib.get('description', '')

        constant = Constant(name, default, domain, symbol, description)

        if self.current_component_type:
            self.current_component_type.add_constant(constant)
        else:
            self.model.add_constant(constant)

    def parse_function(self, node):
        """
        Parses <Function>

        @param node: Node containing the <Function> element
        @type node: xml.etree.Element

        MV: added function for pre and post coupling behavior
        """
        try:
            name = node.lattrib['name']
        except:
            self.raise_error('<Function> must specify a name.')

        try:
            value = node.lattrib['value']
        except:
            self.raise_error("Function '{0}' must have a value.", name)

        dimension = None
        symbol = None
        description = None

        function = Function(name, value, dimension, symbol, description)

        if self.current_component_type:
            self.current_component_type.add_function(function)
        else:
            self.model.add_function(function)


    def parse_data_display(self, node):
        """
        Parses <DataDisplay>

        @param node: Node containing the <DataDisplay> element
        @type node: xml.etree.Element
        """

        if 'title' in node.lattrib:
            title = node.lattrib['title']
        else:
            self.raise_error('<DataDisplay> must have a title.')

        if 'dataregion' in node.lattrib:
            data_region = node.lattrib['dataregion']
        else:
            data_region = None

        self.current_simulation.add_data_display(DataDisplay(title, data_region))

    def parse_data_writer(self, node):
        """
        Parses <DataWriter>

        @param node: Node containing the <DataWriter> element
        @type node: xml.etree.Element
        """

        if 'path' in node.lattrib:
            path = node.lattrib['path']
        else:
            self.raise_error('<DataWriter> must specify a path.')

        if 'filename' in node.lattrib:
            file_path = node.lattrib['filename']
        else:
            self.raise_error("Data writer for '{0}' must specify a filename.",
                             path)

        self.current_simulation.add_data_writer(DataWriter(path, file_path))

    def parse_event_writer(self, node):
        """
        Parses <EventWriter>

        @param node: Node containing the <EventWriter> element
        @type node: xml.etree.Element
        """

        if 'path' in node.lattrib:
            path = node.lattrib['path']
        else:
            self.raise_error('<EventWriter> must specify a path.')

        if 'filename' in node.lattrib:
            file_path = node.lattrib['filename']
        else:
            self.raise_error("Event writer for '{0}' must specify a filename.",
                             path)
                             
        if 'format' in node.lattrib:
            format = node.lattrib['format']
        else:
            self.raise_error("Event writer for '{0}' must specify a format.",
                             path)

        self.current_simulation.add_event_writer(EventWriter(path, file_path, format))

    def parse_derived_parameter(self, node):
        """
        Parses <DerivedParameter>

        @param node: Node containing the <DerivedParameter> element
        @type node: xml.etree.Element
        """

        #if self.current_context.context_type != Context.COMPONENT_TYPE:
        #    self.raise_error('Dynamics must be defined inside a ' +
        #                     'component type')

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('A derived parameter must have a name')

        if 'expression' in node.lattrib:
            expression = node.lattrib['expression']
        else:
            expression = None

        if 'value' in node.lattrib:
            value = node.lattrib['value']
        else:
            value = None

        if 'select' in node.lattrib:
            select = node.lattrib['select']
        else:
            select = None

        self.current_component_type.add_derived_parameter(DerivedParameter(name, value,
                                                                    expression, select))

    def parse_derived_variable(self, node):
        """
        Parses <DerivedVariable>

        @param node: Node containing the <DerivedVariable> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when no name of specified for the derived variable.
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        elif 'exposure' in node.lattrib:
            name = node.lattrib['exposure']
        else:
            self.raise_error('<DerivedVariable> must specify a name')

        params = dict()
        for attr_name in ['dimension', 'exposure', 'select', 'expression', 'reduce', 'required']:
            if attr_name in node.lattrib:
                params[attr_name] = node.lattrib[attr_name]

        self.current_regime.add_derived_variable(DerivedVariable(name, **params))
        
        
    def parse_conditional_derived_variable(self, node):
        """
        Parses <ConditionalDerivedVariable>

        @param node: Node containing the <ConditionalDerivedVariable> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when no name or value is specified for the conditional derived variable.
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        elif 'exposure' in node.lattrib:
            name = node.lattrib['exposure']
        else:
            self.raise_error('<ConditionalDerivedVariable> must specify a name')
            
        if 'exposure' in node.lattrib:
            exposure = node.lattrib['exposure']
        else:
            exposure = None
            
        if 'condition' in node.lattrib:
            condition = node.lattrib['condition']
        else:
            condition = None

        if 'cases' in node.lattrib:
            cases = node.lattrib['cases']
        else:
            cases = None

        conditional_derived_variable = ConditionalDerivedVariable(name, condition, exposure, cases)
        
        self.current_regime.add_conditional_derived_variable(conditional_derived_variable)
        
        self.current_conditional_derived_variable = conditional_derived_variable
        
        self.process_nested_tags(node)
        
        
    def parse_case(self, node):
        """
        Parses <Case>

        @param node: Node containing the <Case> element
        @type node: xml.etree.Element

        @raise ParseError: When no condition or value is specified
        """
        
        try:
            condition = node.lattrib['condition']
        except:
            condition = None
            
        try:
            value = node.lattrib['value']
        except:
            self.raise_error('<Case> must specify a value')

        self.current_conditional_derived_variable.add_case(Case(condition, value))

    def parse_dimension(self, node):
        """
        Parses <Dimension>

        @param node: Node containing the <Dimension> element
        @type node: xml.etree.Element

        @raise ParseError: When the name is not a string or if the
        dimension is not a signed integer.
        """

        try:
            name = node.lattrib['name']
        except:
            self.raise_error('<Dimension> must specify a name')

        description = node.lattrib.get('description', '')

        dim = dict()
        for d in ['l', 'm', 't', 'i', 'k', 'c', 'n']:
            dim[d] = int(node.lattrib.get(d, 0))

        self.model.add_dimension(Dimension(name, description, **dim))

    def parse_dynamics(self, node):
        """
        Parses <Dynamics>

        @param node: Node containing the <Behaviour> element
        @type node: xml.etree.Element
        """

        self.current_dynamics = self.current_component_type.dynamics
        self.current_regime = self.current_dynamics
        self.process_nested_tags(node)
        self.current_regime = None
        self.current_dynamics = None

    def parse_event_connection(self, node):
        """
        Parses <EventConnection>

        @param node: Node containing the <EventConnection> element
        @type node: xml.etree.Element
        """

        if 'from' in node.lattrib:
            from_ = node.lattrib['from']
        else:
            self.raise_error('<EventConnection> must provide a source (from) component reference.')

        if 'to' in node.lattrib:
            to = node.lattrib['to']
        else:
            self.raise_error('<EventConnection> must provide a target (to) component reference.')

        source_port = node.lattrib.get('sourceport', '')
        target_port = node.lattrib.get('targetport', '')
        receiver = node.lattrib.get('receiver', '')
        receiver_container = node.lattrib.get('receivercontainer', '')

        ec = EventConnection(from_, to, source_port, target_port, receiver, receiver_container)
        self.current_structure.add_event_connection(ec)

    def parse_event_out(self, node):
        """
        Parses <EventOut>

        @param node: Node containing the <EventOut> element
        @type node: xml.etree.Element
        """

        try:
            port = node.lattrib['port']
        except:
            self.raise_error('<EventOut> must be specify a port.')

        action = EventOut(port)

        self.current_event_handler.add_action(action)

    def parse_event_port(self, node):
        """
        Parses <EventPort>

        @param node: Node containing the <EventPort> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error(('<EventPort> must specify a name.'))

        if 'direction' in node.lattrib:
            direction = node.lattrib['direction']
        else:
            self.raise_error("Event port '{0}' must specify a direction.")

        direction = direction.lower()
        if direction != 'in' and direction != 'out':
            self.raise_error(('Event port direction must be \'in\' '
                              'or \'out\''))

        description = node.lattrib.get('description', '')
        
        self.current_component_type.add_event_port(EventPort(name, direction, description))

    def parse_exposure(self, node):
        """
        Parses <Exposure>

        @param node: Node containing the <Exposure> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when the exposure name is not
        being defined in the context of a component type.
        """

        if self.current_component_type == None:
            self.raise_error('Exposures must be defined in a component type')

        try:
            name = node.lattrib['name']
        except:
            self.raise_error('<Exposure> must specify a name')

        try:
            choices = node.lattrib['choices']
        except:
            self.raise_error("Exposure '{0}' must specify choices",
                             name)

        try:
            default = node.lattrib['default']
        except:
            self.raise_error("Exposure '{0}' must specify default",
                             name)

        description = node.lattrib.get('description', '')

        self.current_component_type.add_exposure(Exposure(name, choices, default, description))

    def parse_fixed(self, node):
        """
        Parses <Fixed>

        @param node: Node containing the <Fixed> element
        @type node: xml.etree.Element
        """

        try:
            parameter = node.lattrib['parameter']
        except:
            self.raise_error('<Fixed> must specify a parameter to be fixed.')

        try:
            value = node.lattrib['value']
        except:
            self.raise_error("Fixed parameter '{0}'must specify a value.", parameter)

        description = node.lattrib.get('description', '')
        
        self.current_component_type.add_parameter(Fixed(parameter, value, description))

    def parse_for_each(self, node):
        """
        Parses <ForEach>

        @param node: Node containing the <ForEach> element
        @type node: xml.etree.Element
        """
        
        if self.current_structure == None:
            self.raise_error('<ForEach> can only be made within ' +
                             'a structure definition')

        if 'instances' in node.lattrib:
            instances = node.lattrib['instances']
        else:
            self.raise_error('<ForEach> must specify a reference to target'
                             'instances')

        if 'as' in node.lattrib:
            as_ = node.lattrib['as']
        else:
            self.raise_error('<ForEach> must specify a name for the '
                             'enumerated target instances')

        old_structure = self.current_structure
        fe = ForEach(instances, as_)
        self.current_structure.add_for_each(fe)
        self.current_structure = fe
        
        self.process_nested_tags(node)

        self.current_structure = old_structure

    def parse_include(self, node):
        """
        Parses <Include>

        @param node: Node containing the <Include> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when the file to be included is not specified. 
        """
        if not self.include_includes:
            if self.model.debug: print("Ignoring included LEMS file: %s"%node.lattrib['file'])
        else:

            #TODO: remove this hard coding for reading NeuroML includes...
            if 'file' not in node.lattrib:
                if 'href' in node.lattrib:
                    self.model.include_file(node.lattrib['href'], self.include_dirs)
                    return
                else:
                    self.raise_error('<Include> must specify the file to be included.')

            self.model.include_file(node.lattrib['file'], self.include_dirs)

    def parse_kinetic_scheme(self, node):
        """
        Parses <KineticScheme>

        @param node: Node containing the <KineticScheme> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<KineticScheme> must specify a name.')

        if 'nodes' in node.lattrib:
            nodes = node.lattrib['nodes']
        else:
            self.raise_error("Kinetic scheme '{0}' must specify nodes.", name)

        if 'statevariable' in node.lattrib:
            state_variable = node.lattrib['statevariable']
        else:
            self.raise_error("Kinetic scheme '{0}' must specify a state variable.", name)

        if 'edges' in node.lattrib:
            edges = node.lattrib['edges']
        else:
            self.raise_error("Kinetic scheme '{0}' must specify edges.", name)

        if 'edgesource' in node.lattrib:
            edge_source = node.lattrib['edgesource']
        else:
            self.raise_error("Kinetic scheme '{0}' must specify the edge source attribute.", name)

        if 'edgetarget' in node.lattrib:
            edge_target = node.lattrib['edgetarget']
        else:
            self.raise_error("Kinetic scheme '{0}' must specify the edge target attribute.", name)

        if 'forwardrate' in node.lattrib:
            forward_rate = node.lattrib['forwardrate']
        else:
            self.raise_error("Kinetic scheme '{0}' must specify the forward rate attribute.", name)

        if 'reverserate' in node.lattrib:
            reverse_rate = node.lattrib['reverserate']
        else:
            self.raise_error("Kinetic scheme '{0}' must specify the reverse rate attribute", name)

        self.current_regime.add_kinetic_scheme(KineticScheme(name, nodes, state_variable,
                                                             edges, edge_source, edge_target,
                                                             forward_rate, reverse_rate))

    def parse_link(self, node):
        """
        Parses <Link>

        @param node: Node containing the <Link> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<Link> must specify a name')

        if 'type' in node.lattrib:
            type_ = node.lattrib['type']
        else:
            self.raise_error("Link '{0}' must specify a type", name)

        description = node.lattrib.get('description', '')

        self.current_component_type.add_link(Link(name, type_, description))

    def parse_multi_instantiate(self, node):
        """
        Parses <MultiInstantiate>

        @param node: Node containing the <MultiInstantiate> element
        @type node: xml.etree.Element
        """

        if 'component' in node.lattrib:
            component = node.lattrib['component']
        else:
            self.raise_error('<MultiInstantiate> must specify a component reference.')

        if 'number' in node.lattrib:
            number = node.lattrib['number']
        else:
            self.raise_error("Multi instantiation of '{0}' must specify a parameter specifying the number.",
                             component)

        self.current_structure.add_multi_instantiate(MultiInstantiate(component, number))

    def parse_on_condition(self, node):
        """
        Parses <OnCondition>

        @param node: Node containing the <OnCondition> element
        @type node: xml.etree.Element
        """

        try:
            test = node.lattrib['test']
        except:
            self.raise_error('<OnCondition> must specify a test.')
            
        event_handler = OnCondition(test)

        self.current_regime.add_event_handler(event_handler)

        self.current_event_handler = event_handler
        self.process_nested_tags(node)
        self.current_event_handler = None
        
    def parse_on_entry(self, node):
        """
        Parses <OnEntry>

        @param node: Node containing the <OnEntry> element
        @type node: xml.etree.Element
        """

        event_handler = OnEntry()

        self.current_event_handler = event_handler
        self.current_regime.add_event_handler(event_handler)

        self.process_nested_tags(node)

        self.current_event_handler = None

    def parse_on_event(self, node):
        """
        Parses <OnEvent>

        @param node: Node containing the <OnEvent> element
        @type node: xml.etree.Element
        """

        try:
            port = node.lattrib['port']
        except:
            self.raise_error('<OnEvent> must specify a port.')
            
        event_handler = OnEvent(port)

        self.current_regime.add_event_handler(event_handler)

        self.current_event_handler = event_handler
        self.process_nested_tags(node)
        self.current_event_handler = None

    def parse_on_start(self, node):
        """
        Parses <OnStart>

        @param node: Node containing the <OnStart> element
        @type node: xml.etree.Element
        """

        event_handler = OnStart()

        self.current_regime.add_event_handler(event_handler)

        self.current_event_handler = event_handler
        self.process_nested_tags(node)
        self.current_event_handler = None

    def parse_parameter(self, node):
        """
        Parses <Parameter>

        @param node: Node containing the <Parameter> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when the parameter does not have a name.
        @raise ParseError: Raised when the parameter does not have a
        dimension.
        """

        if self.current_component_type == None:
            self.raise_error('Parameters can only be defined in ' +
                             'a component type')

        try:
            name = node.lattrib['name']
        except:
            self.raise_error('<Parameter> must specify a name')

        try:
            dimension = node.lattrib['dimension']
        except:
            self.raise_error("Parameter '{0}' has no dimension",
                             name)

        parameter = Parameter(name, dimension)

        self.current_component_type.add_parameter(parameter)

    def parse_property(self, node):
        """
        Parses <Property>

        @param node: Node containing the <Property> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when the property does not have a name.
        @raise ParseError: Raised when the property does not have a
        dimension.
        """

        if self.current_component_type == None:
            self.raise_error('Property can only be defined in ' +
                             'a component type')

        try:
            name = node.lattrib['name']
        except:
            self.raise_error('<Property> must specify a name')

        try:
            dimension = node.lattrib['dimension']
        except:
            self.raise_error("Property '{0}' has no dimension",
                             name)
                             
        default_value = node.lattrib.get('defaultvalue', None)
        
        property = Property(name, dimension, default_value=default_value)

        self.current_component_type.add_property(property)
        
        
    def parse_index_parameter(self, node):
        """
        Parses <IndexParameter>

        @param node: Node containing the <IndexParameter> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when the IndexParameter does not have a name.
        """

        if self.current_component_type == None:
            self.raise_error('IndexParameters can only be defined in ' +
                             'a component type')

        try:
            name = node.lattrib['name']
        except:
            self.raise_error('<IndexParameter> must specify a name')


        index_parameter = IndexParameter(name)

        self.current_component_type.add_index_parameter(index_parameter)
        
        
    def parse_tunnel(self, node):
        """
        Parses <Tunnel>

        @param node: Node containing the <Tunnel> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when the Tunnel does not have a name.
        """

        try:
            name = node.lattrib['name']
        except:
            self.raise_error('<Tunnel> must specify a name')
        try:
            end_a = node.lattrib['enda']
        except:
            self.raise_error('<Tunnel> must specify: endA')
        try:
            end_b = node.lattrib['enda']
        except:
            self.raise_error('<Tunnel> must specify: endB')
        try:
            component_a = node.lattrib['componenta']
        except:
            self.raise_error('<Tunnel> must specify: componentA')
        try:
            component_b = node.lattrib['componentb']
        except:
            self.raise_error('<Tunnel> must specify: componentB')


        tunnel = Tunnel(name, end_a, end_b, component_a, component_b)

        self.current_structure.add_tunnel(tunnel)
        

    def parse_path(self, node):
        """
        Parses <Path>

        @param node: Node containing the <Path> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<Path> must specify a name.')

        description = node.lattrib.get('description', '')

        self.current_component_type.add_path(Path(name, description))

    def parse_record(self, node):
        """
        Parses <Record>

        @param node: Node containing the <Record> element
        @type node: xml.etree.Element
        """

        if self.current_simulation == None:
            self.raise_error('<Record> must be only be used inside a ' +
                             'simulation specification')

        if 'quantity' in node.lattrib:
            quantity = node.lattrib['quantity']
        else:
            self.raise_error('<Record> must specify a quantity.')

        scale = node.lattrib.get('scale', None)
        color  = node.lattrib.get('color', None)
        id  = node.lattrib.get('id', None)

        self.current_simulation.add_record(Record(quantity, scale, color, id))

    def parse_event_record(self, node):
        """
        Parses <EventRecord>

        @param node: Node containing the <EventRecord> element
        @type node: xml.etree.Element
        """

        if self.current_simulation == None:
            self.raise_error('<EventRecord> must be only be used inside a ' +
                             'simulation specification')

        if 'quantity' in node.lattrib:
            quantity = node.lattrib['quantity']
        else:
            self.raise_error('<EventRecord> must specify a quantity.')

        if 'eventport' in node.lattrib:
            eventPort = node.lattrib['eventport']
        else:
            self.raise_error('<EventRecord> must specify an eventPort.')


        self.current_simulation.add_event_record(EventRecord(quantity, eventPort))

    def parse_regime(self, node):
        """
        Parses <Regime>

        @param node: Node containing the <Behaviour> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            name = ''

        if 'initial' in node.lattrib:
            initial = (node.lattrib['initial'].strip().lower() == 'true')
        else:
            initial = False

        regime = Regime(name, self.current_dynamics, initial)
        old_regime = self.current_regime
        self.current_dynamics.add_regime(regime)
        self.current_regime = regime

        self.process_nested_tags(node)

        self.current_regime = old_regime

    def parse_requirement(self, node):
        """
        Parses <Requirement>

        @param node: Node containing the <Requirement> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<Requirement> must specify a name')

        if 'dimension' in node.lattrib:
            dimension = node.lattrib['dimension']
        else:
            self.raise_error("Requirement \{0}' must specify a dimension.", name)

        self.current_component_type.add_requirement(Requirement(name, dimension))
    
    def parse_component_requirement(self, node):
        """
        Parses <ComponentRequirement>

        @param node: Node containing the <ComponentRequirement> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<ComponentRequirement> must specify a name')

        self.current_component_type.add_component_requirement(ComponentRequirement(name))
    
    def parse_instance_requirement(self, node):
        """
        Parses <InstanceRequirement>

        @param node: Node containing the <InstanceRequirement> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<InstanceRequirement> must specify a name')

        if 'type' in node.lattrib:
            type = node.lattrib['type']
        else:
            self.raise_error("InstanceRequirement \{0}' must specify a type.", name)

        self.current_component_type.add_instance_requirement(InstanceRequirement(name, type))
    
    def parse_run(self, node):
        """
        Parses <Run>

        @param node: Node containing the <Run> element
        @type node: xml.etree.Element
        """

        if 'component' in node.lattrib:
            component = node.lattrib['component']
        else:
            self.raise_error('<Run> must specify a target component')

        if 'variable' in node.lattrib:
            variable = node.lattrib['variable']
        else:
            self.raise_error('<Run> must specify a state variable')

        if 'increment' in node.lattrib:
            increment = node.lattrib['increment']
        else:
            self.raise_error('<Run> must specify an increment for the ' +
                             'state variable')

        if 'total' in node.lattrib:
            total = node.lattrib['total']
        else:
            self.raise_error('<Run> must specify a final value for the ' +
                             'state variable')

        self.current_simulation.add_run(Run(component, variable, increment, total))

    def parse_show(self, node):
        """
        Parses <Show>

        @param node: Node containing the <Show> element
        @type node: xml.etree.Element
        """

        pass

    def parse_simulation(self, node):
        """
        Parses <Simulation>

        @param node: Node containing the <Simulation> element
        @type node: xml.etree.Element
        """

        self.current_simulation = self.current_component_type.simulation

        self.process_nested_tags(node)

        self.current_simulation = None

    def parse_state_assignment(self, node):
        """
        Parses <StateAssignment>

        @param node: Node containing the <StateAssignment> element
        @type node: xml.etree.Element
        """

        if 'variable' in node.lattrib:
            variable = node.lattrib['variable']
        else:
            self.raise_error('<StateAssignment> must specify a variable name')

        if 'value' in node.lattrib:
            value = node.lattrib['value']
        else:
            self.raise_error("State assignment for '{0}' must specify a value.",
                             variable)

        action = StateAssignment(variable, value)

        self.current_event_handler.add_action(action)


    def parse_state_variable(self, node):
        """
        Parses <StateVariable>

        @param node: Node containing the <StateVariable> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when the state variable is not
        being defined in the context of a component type.
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<StateVariable> must specify a name')

        if 'default' in node.lattrib:
            default = node.lattrib['default']
        else:
            self.raise_error("State variable '{0}' must specify a default", name)

        if 'boundaries' in node.lattrib:
            boundaries = node.lattrib['boundaries']
        else:
            boundaries = None

        self.current_regime.add_state_variable(StateVariable(name, default, boundaries))

    def parse_structure(self, node):
        """
        Parses <Structure>

        @param node: Node containing the <Structure> element
        @type node: xml.etree.Element
        """

        self.current_structure = self.current_component_type.structure
        self.process_nested_tags(node)
        self.current_structure = None

    def parse_target(self, node):
        """
        Parses <Target>

        @param node: Node containing the <Target> element
        @type node: xml.etree.Element
        """

        self.model.add_target(node.lattrib['component'])

    def parse_text(self, node):
        """
        Parses <Text>

        @param node: Node containing the <Text> element
        @type node: xml.etree.Element
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<Text> must specify a name.')

        description = node.lattrib.get('description', '')

        self.current_component_type.add_text(Text(name, description))

    def parse_time_derivative(self, node):
        """
        Parses <TimeDerivative>

        @param node: Node containing the <TimeDerivative> element
        @type node: xml.etree.Element

        @raise ParseError: Raised when the time derivative does not hava a variable
        name of a value.
        """

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            self.raise_error('<TimeDerivative> must specify a name.')

        if 'expression' in node.lattrib:
            expression = node.lattrib['expression']
        else:
            self.raise_error("Time derivative for '{0}' must specify an expression.",
                             variable)

        self.current_regime.add_time_derivative(TimeDerivative(name, expression))

    def parse_transition(self, node):
        """
        Parses <Transition>

        @param node: Node containing the <Transition> element
        @type node: xml.etree.Element
        """

        if 'regime' in node.lattrib:
            regime = node.lattrib['regime']
        else:
            self.raise_error('<Transition> mut specify a regime.')

        action = Transition(regime)

        self.current_event_handler.add_action(action)

    def parse_unit(self, node):
        """
        Parses <Unit>

        @param node: Node containing the <Unit> element
        @type node: xml.etree.Element

        @raise ParseError: When the name is not a string or the unit
        specfications are incorrect.

        @raise ModelError: When the unit refers to an undefined dimension.
        """

        try:
            symbol = node.lattrib['symbol']
            dimension = node.lattrib['dimension']
        except:
            self.raise_error('Unit must have a symbol and dimension.')

        if 'power' in node.lattrib:
            power = int(node.lattrib['power'])
        else:
            power = 0

        if 'name' in node.lattrib:
            name = node.lattrib['name']
        else:
            name = ''
            
        if 'scale' in node.lattrib:
            scale = float(node.lattrib['scale'])
        else:
            scale = 1.0
            
        if 'offset' in node.lattrib:
            offset = float(node.lattrib['offset'])
        else:
            offset = 0.0

        self.model.add_unit(Unit(name, symbol, dimension, power, scale, offset))

    def parse_with(self, node):
        """
        Parses <With>

        @param node: Node containing the <With> element
        @type node: xml.etree.Element
        """

        if 'instance' in node.lattrib:
            instance = node.lattrib['instance']
            list = None
            index = None
        elif 'list' in node.lattrib and 'index' in node.lattrib:
            instance = None
            list = node.lattrib['list']
            index = node.lattrib['index']
        else:
            self.raise_error('<With> must specify EITHER instance OR list & index')

        if 'as' in node.lattrib:
            as_ = node.lattrib['as']
        else:
            self.raise_error('<With> must specify a name for the '
                             'target instance')

        self.current_structure.add_with(With(instance, as_, list, index))




"""
Structural properties of component types.

@author: Gautham Ganapathy
@organization: LEMS (http://neuroml.org/lems/, https://github.com/organizations/LEMS)
@contact: gautham@lisphacker.org
"""

from ..base.base import LEMSBase
from ..base.map import Map
from ..base.errors import ModelError

class With(LEMSBase):
    """
    Stores a with-as statement.
    """

    def __init__(self, instance, as_, list=None, index=None):
        """
        Constructor referencing single identified instance, or list/index.
        
        """

        self.instance = instance
        """ Instance to be referenced.
        @type: str """

        self.as_ = as_
        """ Alternative name.
        @type: str """

        self.list = list
        """ list of components, e.g. population
        @type: str """

        self.index = index
        """ index in list to be referenced.
        @type: str """
        

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<With ' + \
          (' instance="{0}"'.format(self.instance) if self.instance else '') +\
          (' list="{0}"'.format(self.list) if self.list else '') + \
          (' index="{0}"'.format(self.index) if self.index else '') + \
          ' as="{1}"/>'.format(self.instance, self.as_)
          
          
class Tunnel(LEMSBase):
    """
    Stores a Tunnel.
    """

    def __init__(self, name, end_a, end_b, component_a, component_b):
        """
        Constructor.
        
        """

        self.name = name
        """ name of Tunnel.
        @type: str """

        self.end_a = end_a
        """ 'A' end of Tunnel.
        @type: str """

        self.end_b = end_b
        """ 'B' end of Tunnel.
        @type: str """

        self.component_a = component_a
        """ Component to create at A.
        @type: str """
        
        self.component_b = component_b
        """ Component to create at B.
        @type: str """

        

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<Tunnel name="{0}"'.format(self.name) + \
               ' endA="{0}"'.format(self.end_a) + \
               ' endB="{0}"'.format(self.end_b) + \
               ' componentA="{0}"'.format(self.component_a) + \
               ' componentB="{0}"'.format(self.component_b) + '/>'

class EventConnection(LEMSBase):
    """
    Stores an event connection specification.
    """

    def __init__(self, from_, to, 
                 source_port, target_port,
                 receiver, receiver_container):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """

        self.from_ = from_
        """ Name of the source component for event.
        @type: str """

        self.to = to
        """ Name of the target component for the event.
        @type: str """

        self.source_port = source_port
        """ Source port name.
        @type: str """

        self.target_port = target_port
        """ Target port name.
        @type: str """

        self.receiver = receiver
        """ Name of the proxy receiver component attached to the target component that actually receiving the event.
        @type: str """

        self.receiver_container = receiver_container
        """ Name of the child component grouping to add the receiver to.
        @type: str """

    def __eq__(self, o):
        return (self.from_ == o.from_ and self.to == o.to and
                self.source_port == o.source_port and self.target_port == o.target_port)


    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<EventConnection' +\
          (' from="{0}"'.format(self.from_) if self.from_ else '') +\
          (' to="{0}"'.format(self.to) if self.to else '') +\
          (' sourcePort="{0}"'.format(self.source_port) if self.source_port else '') +\
          (' targetPort="{0}"'.format(self.target_port) if self.target_port else '') +\
          (' receiver="{0}"'.format(self.receiver) if self.receiver else '') +\
          (' receiverContainer="{0}"'.format(self.receiver_container) if self.receiver_container else '') +\
          '/>'

class ChildInstance(LEMSBase):
    """
    Stores a child instantiation specification.
    """

    def __init__(self, component, referenced_component = None):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """

        self.component = component
        """ Name of the component reference to be used for instantiation.
        @type: str """

        self.referenced_component = referenced_component
        """ Target component being referenced after resolution.
        @type: lems.model.component.FatComponent """
        
    def __eq__(self, o):
        return self.component == o.component

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<ChildInstance component="{0}"/>'.format(self.component)

class Assign(LEMSBase):
    """
    Stores a child assign specification.
    """

    def __init__(self, property, value):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """

        self.property_ = property
        """ Name of the property reference to be used for instantiation.
        @type: str """

        self.value = value
        """ Value of the property.
        @type: str"""

    def __eq__(self, o):
        return self.property_ == o.property_ and self.value == o.value

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """
        return '<Assign property="{0}" value="{1}"/>'.format(self.property_, self.value)


class MultiInstantiate(LEMSBase):
    """
    Stores a child multi-instantiation specification.
    """

    def __init__(self, component=None, number=None, component_type=None):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """
        if component and component_type:
            raise AttributeError("MultiInstantiate should contain either"
                                 " an attribute component or an attribute"
                                 " component_type, not both.")

        self.component = component
        """ Name of the component reference to be used for instantiation.
        @type: str """

        self.component_type = component_type
        """ Name of the component type reference to be used for instantiation.
        @type: str """

        self.number = number
        """ Name of the paramter specifying the number of times the component 
        reference is to be instantiated.
        @type: str"""

        self.assignments = []
        """ List of assignments included in MultiInstantiate.
        @type: list(Assign) """
        
    def __eq__(self, o):
        if self.component:
            flag = self.component == o.component and self.number == o.number
        else:
            flag = self.component_type == o.component_type and self.number == o.number
        return flag

    def add_assign(self, assign):
        """
        Adds an Assign to the structure.

        @param assign: Assign structure.
        @type assign: lems.model.structure.Assign
        """
        self.assignments.append(assign)

    def add(self, child):
        """
        Adds a typed child object to the structure object.

        @param child: Child object to be added.
        """

        if isinstance(child, Assign):
            self.add_assign(child)
        else:
            raise ModelError('Unsupported child element')

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """
        argstr = ''
        if self.component:
            argstr += 'component="{0}" '.format(self.component)
        if self.component_type:
            argstr += 'componentType="{0}" '.format(self.component_type)
        if self.number:
            argstr += 'number="{0}" '.format(self.number)
        if self.assignments:
            chxmlstr = ''
            for assign in self.assignments:
                chxmlstr += assign.toxml()
            return '<MultiInstantiate {0}>{1}</MultiInstantiate>'.format(argstr, chxmlstr)
        else:
            return '<MultiInstantiate {0}/>'.format(argstr)

class ForEach(LEMSBase):
    """
    ForEach specification.
    """
    def __init__(self, instances, as_):

        self.instances = instances
        
        self.as_ = as_
        
        self.event_connections = list()
        """ List of event connections.
        @type: list(lems.model.structure.EventConnection) """
        
        self.for_eachs = list()
        """ List of for each specs.
        @type: list(lems.model.structure.ForEach) """
        
        
    def add_for_each(self, fe):
        """
        Adds a for-each specification.

        @param fe: For-each specification.
        @type fe: lems.model.structure.ForEach
        """

        self.for_eachs.append(fe)
        
        
    def add_event_connection(self, ec):
        """
        Adds an event conenction to the structure.

        @param ec: Event connection.
        @type ec: lems.model.structure.EventConnection
        """

        self.event_connections.append(ec)
        
    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """
        chxmlstr = ''

        for event_connection in self.event_connections:
            chxmlstr += event_connection.toxml()

        for for_each in self.for_eachs:
            chxmlstr += for_each.toxml()


        return '<ForEach instances="{0}" as="{1}">{2}</ForEach>'.format(self.instances, self.as_, chxmlstr)
        
        
class Structure(LEMSBase):
    """
    Stores structural properties of a component type.
    """

    def __init__(self):
        """
        Constructor.
        """

        self.withs = Map()
        """ Map of With statements.
        @type: Map(str -> lems.model.structure.With) """
        
        self.tunnels = Map()
        """ Map of tunnel statements.
        @type: Map(str -> lems.model.structure.Tunnel) """

        self.event_connections = list()
        """ List of event connections.
        @type: list(lems.model.structure.EventConnection) """

        self.child_instances = list()
        """ List of child instantations.
        @type: list(lems.model.structure.ChildInstance) """

        self.multi_instantiates = list()
        """ List of child multi-instantiations.
        @type: list(lems.model.structure.MultiInstantiate) """

        self.for_eachs = list()
        """ List of for each specs.
        @type: list(lems.model.structure.ForEach) """
        
    def has_content(self):
        if len(self.withs)==0 and \
           len(self.event_connections)==0 and \
           len(self.child_instances)==0 and \
           len(self.multi_instantiates)==0 and \
           len(self.for_eachs)==0:
               return False
        else:
            return True

    def add_with(self, with_):
        """
        Adds a with-as specification to the structure.

        @param with_: With-as specification.
        @type with_: lems.model.structure.With
        """

        self.withs[with_.as_] = with_

    def add_tunnel(self, tunnel):
        """
        Adds a tunnel specification to the structure.

        @param tunnel: tunnel specification.
        @type tunnel: lems.model.structure.Tunnel
        """

        self.tunnels[tunnel.name] = tunnel
        
    def add_event_connection(self, ec):
        """
        Adds an event conenction to the structure.

        @param ec: Event connection.
        @type ec: lems.model.structure.EventConnection
        """

        self.event_connections.append(ec)

    def add_child_instance(self, ci):
        """
        Adds a child instantiation specification.

        @param ci: Child instantiation specification.
        @type ci: lems.model.structure.ChildInstance
        """

        self.child_instances.append(ci)

    def add_multi_instantiate(self, mi):
        """
        Adds a child multi-instantiation specification.

        @param mi: Child multi-instantiation specification.
        @type mi: lems.model.structure.MultiInstantiate
        """

        self.multi_instantiates.append(mi)

    def add_for_each(self, fe):
        """
        Adds a for-each specification.

        @param fe: For-each specification.
        @type fe: lems.model.structure.ForEach
        """

        self.for_eachs.append(fe)

    def add(self, child):
        """
        Adds a typed child object to the structure object.

        @param child: Child object to be added.
        """

        if isinstance(child, With):
            self.add_with(child)
        elif isinstance(child, EventConnection):
            self.add_event_connection(child)
        elif isinstance(child, ChildInstance):
            self.add_child_instance(child)
        elif isinstance(child, MultiInstantiate):
            self.add_multi_instantiate(child)
        elif isinstance(child, ForEach):
            self.add_for_each(child)
        else:
            raise ModelError('Unsupported child element')

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """
        chxmlstr = ''

        for with_ in self.withs:
            chxmlstr += with_.toxml()

        for event_connection in self.event_connections:
            chxmlstr += event_connection.toxml()

        for child_instance in self.child_instances:
            chxmlstr += child_instance.toxml()

        for multi_instantiate in self.multi_instantiates:
            chxmlstr += multi_instantiate.toxml()

        for for_each in self.for_eachs:
            chxmlstr += for_each.toxml()

        if chxmlstr:
            xmlstr = '<Structure>' + chxmlstr + '</Structure>'
        else:
            xmlstr = ''

        return xmlstr



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
An interactive power spectra plot generated from a TVB TimeSeries datatype.

Usage
::

    #Load the demo data
    import numpy
    data = numpy.load("demos/demo_data_region_16s_2048Hz.npy")
    period = 0.00048828125 #NOTE: Providing period in seconds

    #Create a tvb TimeSeries object
    import tvb.datatypes.time_series
    tsr = tvb.datatypes.time_series.TimeSeriesRegion()
    tsr.data = data
    tsr.sample_period = period

    #Create and launch the interactive visualiser
    import tvb.simulator.power_spectra_interactive as ps_int
    psi = ps_int.PowerSpectraInteractive(time_series=tsr)
    psi.show()


.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import numpy
import pylab
import matplotlib.widgets as widgets
from tvb.simulator.common import get_logger
import tvb.datatypes.time_series as time_series_datatypes
from tvb.basic.neotraits.api import HasTraits, Attr, Int

LOG = get_logger(__name__)

# Define a colour theme... see: matplotlib.colors.cnames.keys()
BACKGROUNDCOLOUR = "slategrey"
EDGECOLOUR = "darkslateblue"
AXCOLOUR = "steelblue"
BUTTONCOLOUR = "steelblue"
HOVERCOLOUR = "blue"


class PowerSpectraInteractive(HasTraits):
    """
    The graphical interface for visualising the power-spectra (FFT) of a
    timeseries provide controls for setting:

        - which state-variable and mode to display [sets]
        - log or linear scaling for the power or frequency axis [binary]
        - sementation lenth [set]
        - windowing function [set]
        - power normalisation [binary] (emphasise relative frequency contribution)
        - show std or sem [binary]


    """

    time_series = Attr(
        field_type=time_series_datatypes.TimeSeries,
        label="Timeseries",
        default=None,
        doc=""" The timeseries to which the FFT is to be applied.""")

    first_n = Int(
        label="Display the first 'n'",
        default=-1,
        doc="""Primarily intended for displaying the first N components of a 
            surface PCA timeseries. Defaults to -1, meaning it'll display all
            of 'space' (ie, regions or vertices or channels). In other words,
            for Region or M/EEG timeseries you can ignore this, but, for a 
            surface timeseries it really must be set.""")


    def __init__(self, **kwargs):
        """
        Initialise based on provided keywords or their traited defaults. Also,
        initialise the place-holder attributes that aren't filled until the
        show() method is called.

        """
        #figure
        self.ifft_fig = None

        #time-series
        self.fft_ax = None

        #Current state
        self.xscale = "linear"
        self.yscale = "log"
        self.mode = 0
        self.variable = 0
        self.show_sem = False
        self.show_std = False
        self.normalise_power = "no"
        self.window_length = 0.25
        self.window_function = "None"

        #Selectors
        self.xscale_selector = None
        self.yscale_selector = None
        self.mode_selector = None
        self.variable_selector = None
        self.show_sem_selector = None
        self.show_std_selector = None
        self.normalise_power_selector = None
        self.window_length_selector = None
        self.window_function_selector = None

        #
        possible_freq_steps = [2**x for x in range(-2, 7)] #Hz
        #possible_freq_steps.append(1.0 / self.time_series_length) #Hz
        self.possible_window_lengths = 1.0 / numpy.array(possible_freq_steps) #s
        self.freq_step = 1.0 / self.window_length
        self.frequency = None
        self.spectra = None
        self.spectra_norm = None

        #Sliders
        #self.window_length_slider = None


    def configure(self):
        """ Seperate configure cause ttraits be busted... """
        LOG.debug("time_series shape: %s" % str(self.time_series.data.shape))
        #TODO: if isinstance(self.time_series, TimeSeriesSurface) and self.first_n == -1: #LOG.error, return.
        self.data = self.time_series.data[:, :, :self.first_n, :]
        self.period = self.time_series.sample_period
        self.max_freq = 0.5 / self.period
        self.units = "Hz"
        self.tpts = self.data.shape[0]
        self.nsrs = self.data.shape[2]
        self.time_series_length = self.tpts * self.period
        self.time = numpy.arange(self.tpts) * self.period
        self.labels = ["channel_%0.3d" % k for k in range(self.nsrs)]


    def show(self):
        """ Generate the interactive power-spectra figure. """
        #Make sure everything is configured
        self.configure()

        #Make the figure:
        self.create_figure()

        #Selectors
        self.add_xscale_selector()
        self.add_yscale_selector()
        self.add_mode_selector()
        self.add_variable_selector()
        self.add_normalise_power_selector()
        self.add_window_length_selector()
        self.add_window_function_selector()

        #Sliders
        #self.add_window_length_slider() #Want discrete values
        #self.add_scaling_slider()

        #...
        self.calc_fft()

        #Plot timeseries
        self.plot_spectra()

        pylab.show()

    ##------------------------------------------------------------------------##
    ##------------------ Functions for building the figure -------------------##
    ##------------------------------------------------------------------------##
    def create_figure(self):
        """ Create the figure and time-series axes. """
        time_series_type = self.time_series.__class__.__name__
        try:
            figure_window_title = "Interactive power spectra: " + time_series_type
            pylab.close(figure_window_title)
            self.ifft_fig = pylab.figure(num = figure_window_title,
                                        figsize = (16, 8),
                                        facecolor = BACKGROUNDCOLOUR, 
                                        edgecolor = EDGECOLOUR)
        except ValueError:
            LOG.info("My life would be easier if you'd update your PyLab...")
            figure_number = 42
            pylab.close(figure_number)
            self.ifft_fig = pylab.figure(num = figure_number, 
                                         figsize = (16, 8), 
                                         facecolor = BACKGROUNDCOLOUR, 
                                         edgecolor = EDGECOLOUR)

        self.fft_ax = self.ifft_fig.add_axes([0.15, 0.2, 0.7, 0.75])


    def add_xscale_selector(self):
        """
        Add a radio button to the figure for selecting which scaling the x-axes
        should use.
        """
        pos_shp = [0.45, 0.02, 0.05, 0.104]
        rax = self.ifft_fig.add_axes(pos_shp, facecolor=AXCOLOUR, title="xscale")
        xscale_tuple = ("log", "linear")
        self.xscale_selector = widgets.RadioButtons(rax, xscale_tuple, active=1)
        self.xscale_selector.on_clicked(self.update_xscale)


    def add_yscale_selector(self):
        """
        Add a radio button to the figure for selecting which scaling the y-axes
        should use.
        """
        pos_shp = [0.02, 0.5, 0.05, 0.104]
        rax = self.ifft_fig.add_axes(pos_shp, facecolor=AXCOLOUR, title="yscale")
        yscale_tuple = ("log", "linear")
        self.yscale_selector = widgets.RadioButtons(rax, yscale_tuple, active=0)
        self.yscale_selector.on_clicked(self.update_yscale)


    def add_mode_selector(self):
        """
        Add a radio button to the figure for selecting which mode of the model
        should be displayed.
        """
        pos_shp = [0.02, 0.07, 0.05, 0.1+0.002*self.data.shape[3]]
        rax = self.ifft_fig.add_axes(pos_shp, facecolor=AXCOLOUR, title="Mode")
        mode_tuple = tuple(range(self.data.shape[3]))
        self.mode_selector = widgets.RadioButtons(rax, mode_tuple, active=0)
        self.mode_selector.on_clicked(self.update_mode)


    def add_variable_selector(self):
        """
        Generate radio selector buttons to set which state variable is 
        displayed.
        """
        noc = self.data.shape[1] # number of choices
        #State variable for the x axis
        pos_shp = [0.02, 0.22, 0.05, 0.12+0.008*noc]
        rax = self.ifft_fig.add_axes(pos_shp, facecolor=AXCOLOUR, 
                                     title="state variable")
        self.variable_selector = widgets.RadioButtons(rax, tuple(range(noc)), 
                                                    active=0)
        self.variable_selector.on_clicked(self.update_variable)


    def add_window_length_selector(self):
        """
        Generate radio selector buttons to set the window length is seconds.
        """
        noc = self.possible_window_lengths.shape[0] # number of choices
        #State variable for the x axis
        pos_shp = [0.88, 0.07, 0.1, 0.12+0.02*noc]
        rax = self.ifft_fig.add_axes(pos_shp, facecolor=AXCOLOUR, 
                                     title="Segment length")
        wl_tup = tuple(self.possible_window_lengths)
        self.window_length_selector = widgets.RadioButtons(rax, wl_tup, active=4)
        self.window_length_selector.on_clicked(self.update_window_length)


    def add_window_function_selector(self):
        """
        Generate radio selector buttons to set the windowing function.
        """
        #TODO: add support for kaiser, requiers specification of beta.
        wf_tup = ("None", "hamming", "bartlett", "blackman", "hanning") 
        noc = len(wf_tup) # number of choices
        #State variable for the x axis
        pos_shp = [0.88, 0.77, 0.085, 0.12+0.01*noc]
        rax = self.ifft_fig.add_axes(pos_shp, facecolor=AXCOLOUR, 
                                    title="Windowing function")
        self.window_function_selector = widgets.RadioButtons(rax, wf_tup, active=0)
        self.window_function_selector.on_clicked(self.update_window_function)


    def add_normalise_power_selector(self):
        """
        Add a radio button to chose whether or not the power of all spectra 
        shouold be normalised to 1.
        """
        pos_shp = [0.02, 0.8, 0.05, 0.104]
        rax = self.ifft_fig.add_axes(pos_shp, facecolor=AXCOLOUR, title="normalise")
        np_tuple = ("yes", "no")
        self.normalise_power_selector = widgets.RadioButtons(rax, np_tuple, active=1)
        self.normalise_power_selector.on_clicked(self.update_normalise_power)


    ##------------------------------------------------------------------------##
    ##------------------ Functions for updating the state --------------------##
    ##------------------------------------------------------------------------##
    def calc_fft(self):
        """
        Calculate FFT using current state of the window_length, window_function,
        """
        #Segment time-series, overlapping if necessary
        nseg = int(numpy.ceil(self.time_series_length / self.window_length))
        if nseg != 1:
            seg_tpts = self.window_length / self.period
            overlap = ((seg_tpts * nseg) - self.tpts) / (nseg-1)
            starts = [max(seg*(seg_tpts - overlap), 0) for seg in range(nseg)]
            segments = [self.data[start:start+seg_tpts] for start in starts] 
            segments = [segment[:, :, :, numpy.newaxis] for segment in segments]
            time_series = numpy.concatenate(segments, axis=4)
        else:
            time_series = self.data[:, :, :, :, numpy.newaxis]
            seg_tpts = time_series.shape[0]

        #Base-line correct segmented time-series
        time_series = time_series - time_series.mean(axis=0)[numpy.newaxis, :]

        #Apply windowing function
        if self.window_function != "None":
            window_function = eval("".join(("numpy.", self.window_function)))
            window_mask = numpy.reshape(window_function(seg_tpts), 
                                        (seg_tpts, 1, 1, 1, 1))
            time_series = time_series * window_mask

        #Calculate the FFT
        result =  numpy.fft.fft(time_series, axis=0)
        nfreq = len(result)/2

        self.frequency = numpy.arange(0, self.max_freq, self.freq_step)
        LOG.debug("frequency shape: %s" % str(self.frequency.shape))

        self.spectra = numpy.mean(numpy.abs(result[1:nfreq+1])**2, axis=-1)
        LOG.debug("spectra shape: %s" % str(self.spectra.shape))

        self.spectra_norm = (self.spectra / numpy.sum(self.spectra, axis=0))
        LOG.debug("spectra_norm shape: %s" % str(self.spectra_norm.shape))

        #import pdb; pdb.set_trace()
#        self.spectra_std = numpy.std(numpy.abs(result[:nfreq]), axis=4)
#        self.spectra_sem = self.spectra_std / time_series.shape[4]

    ##------------------------------------------------------------------------##
    ##------------------ Functions for updating the figure -------------------##
    ##------------------------------------------------------------------------##
    def update_xscale(self, xscale):
        """ 
        Update the FFT axes' xscale to either log or linear based on radio
        button selection.
        """
        self.xscale = xscale
        self.fft_ax.set_xscale(self.xscale)
        pylab.draw()


    def update_yscale(self, yscale):
        """ 
        Update the FFT axes' yscale to either log or linear based on radio
        button selection.
        """
        self.yscale = yscale
        self.fft_ax.set_yscale(self.yscale)
        pylab.draw()


    def update_mode(self, mode):
        """ Update the visualised mode based on radio button selection. """
        self.mode = mode
        self.plot_spectra()


    def update_variable(self, variable):
        """ 
        Update state variable being plotted based on radio buttton selection.
        """
        self.variable = variable
        self.plot_spectra()


    def update_normalise_power(self, normalise_power):
        """ Update whether to normalise based on radio button selection. """
        self.normalise_power = normalise_power
        self.plot_spectra()


    def update_window_length(self, length):
        """
        Update timeseries window length based on the selected value.
        """
        #TODO: need this casting but not sure why, don't need int() with mode...
        self.window_length = numpy.float64(length) 
        #import pdb; pdb.set_trace()
        self.freq_step = 1.0 / self.window_length
        self.update_spectra()


    def update_window_function(self, window_function):
        """
        Update windowing function based on the radio button selection.
        """
        self.window_function = window_function
        self.update_spectra()


    def update_spectra(self):
        """ Clear the axes and redraw the power-spectra. """
        self.calc_fft()
        self.plot_spectra()


#    def plot_std(self):
#        """ Plot """
#        std = (self.spectra[:, self.variable, :, self.mode] +
#               self.spectra_std[:, self.variable, :, self.mode])
#        self.fft_ax.plot(self.frequency, std, "--")
#    
#    
#    def plot_sem(self):
#        """  """
#        sem = (self.spectra[:, self.variable, :, self.mode] +
#               self.spectra_sem[:, self.variable, :, self.mode])
#        self.fft_ax.plot(self.frequency, sem, ":")


    def plot_spectra(self):
        """ Plot the power spectra. """
        self.fft_ax.clear()
        # Set title and axis labels
        time_series_type = self.time_series.__class__.__name__
        self.fft_ax.set(title = time_series_type)
        self.fft_ax.set(xlabel = "Frequency (%s)" % self.units)
        self.fft_ax.set(ylabel = "Power")

        # Set x and y scale based on curent radio button selection.
        self.fft_ax.set_xscale(self.xscale)
        self.fft_ax.set_yscale(self.yscale)

        if hasattr(self.fft_ax, 'autoscale'):
            self.fft_ax.autoscale(enable=True, axis='both', tight=True)

        #import pdb; pdb.set_trace()
        #Plot the power spectra
        if self.normalise_power == "yes":
            self.fft_ax.plot(self.frequency, 
                             self.spectra_norm[:, self.variable, :, self.mode])
        else:
            self.fft_ax.plot(self.frequency, 
                             self.spectra[:, self.variable, :, self.mode])

#        #TODO: Need to ensure colour matching... and allow region selection.
#        #If requested, add standard deviation
#        if self.show_std:
#            self.plot_std(self)
#            
#        #If requested, add standard error in mean
#        if self.show_sem:
#            self.plot_sem(self)

        pylab.draw()


if __name__ == "__main__":
    # Do some stuff that tests or makes use of this module...
    LOG.info("Testing %s module..." % __file__)
    try:
        data = numpy.load("../demos/demo_data_region_16s_2048Hz.npy")
    except IOError:
        LOG.error("Can't load demo data. Run demos/generate_region_demo_data.py")
        raise

    period = 0.00048828125 #NOTE: Providing period in seconds
    tsr = time_series_datatypes.TimeSeriesRegion()
    tsr.data = data
    tsr.sample_period = period

    psi = PowerSpectraInteractive(time_series=tsr)
    psi.show()




# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)

"""
A collection of neuronal dynamics models.

Specific models inherit from the abstract class Model.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>
.. moduleauthor:: Paula Sanz Leon <Paula@tvb.invalid>
.. moduleauthor:: Gaurav Malhotra <Gaurav@tvb.invalid>
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""

# we don't import all models by default here, since they are time consuming
# to setup (e.g. numba gufunc compilation), but provide them as module-level
# properties for compatibility with previous version of TVB. For example
#
#     import tvb.simulator.models.Epileptor
#
# works, but only lazily loads the tvb.simulator.models.epileptor module
# and returns the Epileptor class.

from enum import Enum


class ModelsEnum(Enum):
    BASE_MODEL = "Model"
    EPILEPTOR = "Epileptor"
    EPILEPTOR_2D = "Epileptor2D"
    EPILEPTOR_RS = "EpileptorRestingState"
    EPILEPTOR_CODIM_3 = "EpileptorCodim3"
    EPILEPTOR_CODIM_3_SLOW = "EpileptorCodim3SlowMod"
    HOPFIELD = "Hopfield"
    JANSEN_RIT = "JansenRit"
    ZETTERBERG_JANSEN = "ZetterbergJansen"
    LARTER_BREAKSPEAR = "LarterBreakspear"
    LINEAR = "Linear"
    GENERIC_2D_OSCILLATOR = "Generic2dOscillator"
    KURAMOTO = "Kuramoto"
    SUP_HOPF = "SupHopf"
    REDUCED_SET_FITZ_HUGH_NAGUMO = "ReducedSetFitzHughNagumo"
    REDUCED_SET_HINDMARSH_ROSE = "ReducedSetHindmarshRose"
    WILSON_COWAN = "WilsonCowan"
    REDUCED_WONG_WANG = "ReducedWongWang"
    REDUCED_WONG_WANG_EXCH_INH = "ReducedWongWangExcInh"
    ZERLAUT_FIRST_ORDER = "ZerlautAdaptationFirstOrder"
    ZERLAUT_SECOND_ORDER = "ZerlautAdaptationSecondOrder"
    MONTBRIO_PAZO_ROXIN = "MontbrioPazoRoxin"
    COOMBES_BYRNE = "CoombesByrne"
    COOMBES_BYRNE_2D = "CoombesByrne2D"
    GAST_SCHMIDT_KNOSCHE_SD = "GastSchmidtKnosche_SD"
    GAST_SCHMIDT_KNOSCHE_SF = "GastSchmidtKnosche_SF"
    DUMONT_GUTKIN = "DumontGutkin"

    def get_class(self):
        return _get_imported_model(self.value)

    @staticmethod
    def get_base_model_subclasses():
        return [model.get_class() for model in list(ModelsEnum) if model != ModelsEnum.BASE_MODEL]


def _get_imported_model(model):
    import sys
    # Imported modules
    imported_modules = sys.modules['tvb.simulator.models']
    try:
        return getattr(imported_modules, model)
    except AttributeError:
        return None


_module_models = {
    'base': [ModelsEnum.BASE_MODEL],
    'epileptor': [ModelsEnum.EPILEPTOR, ModelsEnum.EPILEPTOR_2D],
    'epileptor_rs': [ModelsEnum.EPILEPTOR_RS],
    'epileptorcodim3': [ModelsEnum.EPILEPTOR_CODIM_3, ModelsEnum.EPILEPTOR_CODIM_3_SLOW],
    'hopfield': [ModelsEnum.HOPFIELD],
    'jansen_rit': [ModelsEnum.JANSEN_RIT, ModelsEnum.ZETTERBERG_JANSEN],
    'larter_breakspear': [ModelsEnum.LARTER_BREAKSPEAR],
    'linear': [ModelsEnum.LINEAR],
    'oscillator': [ModelsEnum.GENERIC_2D_OSCILLATOR, ModelsEnum.KURAMOTO, ModelsEnum.SUP_HOPF],
    'stefanescu_jirsa': [ModelsEnum.REDUCED_SET_HINDMARSH_ROSE, ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO],
    'wilson_cowan': [ModelsEnum.WILSON_COWAN],
    'wong_wang': [ModelsEnum.REDUCED_WONG_WANG],
    'wong_wang_exc_inh': [ModelsEnum.REDUCED_WONG_WANG_EXCH_INH],
    'zerlaut': [ModelsEnum.ZERLAUT_FIRST_ORDER, ModelsEnum.ZERLAUT_SECOND_ORDER],
    'infinite_theta': [ModelsEnum.MONTBRIO_PAZO_ROXIN, ModelsEnum.COOMBES_BYRNE, ModelsEnum.COOMBES_BYRNE_2D, ModelsEnum.GAST_SCHMIDT_KNOSCHE_SF, ModelsEnum.GAST_SCHMIDT_KNOSCHE_SD, ModelsEnum.DUMONT_GUTKIN],
}


def _find_lems_models():
    import os, tvb.dsl
    dsl_path = os.path.dirname(os.path.abspath(tvb.dsl.__file__))
    xml_folder = os.path.join(dsl_path, 'NeuroML', 'XMLmodels')
    for xml_filename in os.listdir(xml_folder):
        # try to get model name
        fullfilename = os.path.join(xml_folder, xml_filename)
        print(fullfilename)
        with open(fullfilename, 'r') as fd:
            for line in fd.readlines():
                if '<ComponentType name=' in line:
                    _, name_, *_ = line.strip().split()
                    _, name, _ = name_.split("=")[1].split('"')
                    break


def _delay_import_one(mod, model):
    """Create getter thunk for module and model name.
    """
    import importlib
    def do_import(_):
        module_name = f'tvb.simulator.models.{mod}'
        model_module = importlib.import_module(module_name)
        return getattr(model_module, model)
    return property(do_import)


def _delay_model_imports():
    """Set up this module with all properties for all models.
    """
    # create substitute module class & object
    class _Module:
        pass
    module = _Module()
    module.__dict__ = globals()
    # create properties for each model
    for mod, models in _module_models.items():
        for model in models:
            setattr(_Module, model.value, _delay_import_one(mod, model.value))
    # register module object
    import sys
    module._module = sys.modules[module.__name__]
    module._pmodule = module
    sys.modules[module.__name__] = module


_delay_model_imports()

